# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import pickle
import subprocess

from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored.common import vim_utils
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored.infra_drivers.openstack import openstack
from tacker.sol_refactored.nfvo import nfvo_client
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields as v2fields


LOG = logging.getLogger(__name__)

CONF = config.CONF


# common sub method
def _get_obj_by_id(obj_id, obj_array):
    # This method assumes that an object that id is obj_id exists.
    for obj in obj_array:
        if obj.id == obj_id:
            return obj
    # not reach here


# sub method for making id
def _make_combination_id(a, b):
    return '{}-{}'.format(a, b)


class VnfLcmDriverV2(object):

    def __init__(self):
        self.endpoint = CONF.v2_vnfm.endpoint
        self.nfvo_client = nfvo_client.NfvoClient()

    def grant(self, context, lcmocc, inst, vnfd):
        # grant exchange
        # NOTE: the api_version of NFVO supposes 1.4.0 at the moment.

        # make common part of grant_req among operations
        grant_req = objects.GrantRequestV1(
            vnfInstanceId=inst.id,
            vnfLcmOpOccId=lcmocc.id,
            vnfdId=inst.vnfdId,
            operation=lcmocc.operation,
            isAutomaticInvocation=lcmocc.isAutomaticInvocation
        )
        grant_req._links = objects.GrantRequestV1_Links(
            vnfLcmOpOcc=objects.Link(
                href=lcmocc_utils.lcmocc_href(lcmocc.id, self.endpoint)),
            vnfInstance=objects.Link(
                href=inst_utils.inst_href(inst.id, self.endpoint)))

        # make operation specific part of grant_req and check request
        # parameters if necessary.
        method = getattr(self, "%s_%s" % (lcmocc.operation.lower(), 'grant'))
        method(grant_req, lcmocc.operationParams, inst, vnfd)

        # NOTE: if not granted, 403 error raised.
        grant = self.nfvo_client.grant(context, grant_req)

        return grant_req, grant

    def post_grant(self, context, lcmocc, inst, grant_req, grant, vnfd):
        method = getattr(self,
                         "%s_%s" % (lcmocc.operation.lower(), 'post_grant'),
                         None)
        if method:
            method(context, lcmocc, inst, grant_req, grant, vnfd)

    def _exec_mgmt_driver_script(self, operation, flavour_id, req, inst,
            grant_req, grant, vnfd):
        script = vnfd.get_interface_script(flavour_id, operation)
        if script is None:
            return

        tmp_csar_dir = vnfd.make_tmp_csar_dir()
        script_dict = {
            'operation': operation,
            'request': req.to_dict(),
            'vnf_instance': inst.to_dict(),
            'grant_request': (grant_req.to_dict()
                              if grant_req is not None else None),
            'grant_response': (grant.to_dict()
                               if grant is not None else None),
            'tmp_csar_dir': tmp_csar_dir
        }
        # script is relative path to Definitions/xxx.yaml
        script_path = os.path.join(tmp_csar_dir, "Definitions", script)

        out = subprocess.run(["python3", script_path],
            input=pickle.dumps(script_dict),
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        vnfd.remove_tmp_csar_dir(tmp_csar_dir)

        if out.returncode != 0:
            LOG.debug("execute %s failed: %s", operation, out.stderr)
            msg = "{} failed: {}".format(operation, out.stderr)
            raise sol_ex.MgmtDriverExecutionFailed(sol_detail=msg)

        LOG.debug("execute %s of %s success.", operation, script)

    def _make_inst_info_common(self, lcmocc, inst_saved, inst, vnfd):
        if not inst.obj_attr_is_set('instantiatedVnfInfo'):
            # NOTE: This is only the case that operation is MODIFY_INFO
            # and vnf instance is never instantiated after creation.
            return

        # make vim independent part of instantiatedVnfInfo.
        # scaleStatus and maxScaleLevels at the moment.
        inst_info = inst.instantiatedVnfInfo
        req = lcmocc.operationParams

        if lcmocc.operation == v2fields.LcmOperationType.INSTANTIATE:
            # create scaleStatus and maxScaleLevels
            flavour_id = req.flavourId
            if req.obj_attr_is_set('instantiationLevelId'):
                inst_level = req.instantiationLevelId
            else:
                inst_level = vnfd.get_default_instantiation_level(flavour_id)

            # make scaleStatus from tosca.policies.nfv.InstantiationLevels
            # definition.
            scale_info = vnfd.get_scale_info_from_inst_level(flavour_id,
                                                             inst_level)
            scale_status = [
                objects.ScaleInfoV2(
                    aspectId=aspect_id,
                    scaleLevel=value['scale_level']
                )
                for aspect_id, value in scale_info.items()
            ]
            max_scale_levels = [
                objects.ScaleInfoV2(
                    aspectId=obj.aspectId,
                    scaleLevel=vnfd.get_max_scale_level(flavour_id,
                                                        obj.aspectId)
                )
                for obj in scale_status
            ]

            if scale_status:
                inst_info.scaleStatus = scale_status
                inst_info.maxScaleLevels = max_scale_levels

            if req.obj_attr_is_set('localizationLanguage'):
                inst_info.localizationLanguage = req.localizationLanguage
        elif lcmocc.operation != v2fields.LcmOperationType.TERMINATE:
            inst_info_saved = inst_saved.instantiatedVnfInfo
            if inst_info_saved.obj_attr_is_set('scaleStatus'):
                inst_info.scaleStatus = inst_info_saved.scaleStatus
                inst_info.maxScaleLevels = inst_info_saved.maxScaleLevels
            if inst_info_saved.obj_attr_is_set('localizationLanguage'):
                inst_info.localizationLanguage = (
                    inst_info_saved.localizationLanguage)

        if lcmocc.operation == v2fields.LcmOperationType.SCALE:
            # adjust scaleStatus
            num_steps = req.numberOfSteps
            if req.type == 'SCALE_IN':
                num_steps *= -1

            for aspect_info in inst_info.scaleStatus:
                if aspect_info.aspectId == req.aspectId:
                    aspect_info.scaleLevel += num_steps
                    break

    def _script_method_name(self, operation):
        # According to the definition in etsi_nfv_sol001_vnfd_types.yaml,
        # get script method name from lcmocc.operation.
        # only MODIFY_INFO and CHANGE_EXT_CONN are exceptional.
        if operation == v2fields.LcmOperationType.MODIFY_INFO:
            return "modify_information"
        elif operation == v2fields.LcmOperationType.CHANGE_EXT_CONN:
            return "change_external_connectivity"
        else:
            return operation.lower()

    def process(self, context, lcmocc, inst, grant_req, grant, vnfd):
        # save inst to use updating lcmocc after process done
        inst_saved = inst.obj_clone()

        # perform preamble LCM script
        req = lcmocc.operationParams
        operation = "{}_start".format(
            self._script_method_name(lcmocc.operation))
        if lcmocc.operation == v2fields.LcmOperationType.INSTANTIATE:
            flavour_id = req.flavourId
        elif inst.obj_attr_is_set('instantiatedVnfInfo'):
            flavour_id = inst.instantiatedVnfInfo.flavourId
        else:
            # NOTE: This is only the case that operation is MODIFY_INFO
            # and vnf instance is never instantiated after creation.
            flavour_id = None
        self._exec_mgmt_driver_script(operation,
            flavour_id, req, inst, grant_req, grant, vnfd)

        # main process
        method = getattr(self, "%s_%s" % (lcmocc.operation.lower(), 'process'))
        method(context, lcmocc, inst, grant_req, grant, vnfd)

        # perform postamble LCM script
        operation = "{}_end".format(self._script_method_name(lcmocc.operation))
        self._exec_mgmt_driver_script(operation,
            flavour_id, req, inst, grant_req, grant, vnfd)

        self._make_inst_info_common(lcmocc, inst_saved, inst, vnfd)
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

    def rollback(self, context, lcmocc, inst, grant_req, grant, vnfd):
        method = getattr(self,
                         "%s_%s" % (lcmocc.operation.lower(), 'rollback'),
                         None)
        if method:
            method(context, lcmocc, inst, grant_req, grant, vnfd)
        else:
            raise sol_ex.RollbackNotSupported(op=lcmocc.operation)

    def _get_link_ports(self, inst_req):
        names = set()
        if inst_req.obj_attr_is_set('extVirtualLinks'):
            for ext_vl in inst_req.extVirtualLinks:
                for ext_cp in ext_vl.extCps:
                    for cp_config in ext_cp.cpConfig.values():
                        if cp_config.obj_attr_is_set('linkPortId'):
                            names.add(ext_cp.cpdId)

        if inst_req.obj_attr_is_set('extManagedVirtualLinks'):
            for ext_mgd_vl in inst_req.extManagedVirtualLinks:
                if ext_mgd_vl.obj_attr_is_set('vnfLinkPort'):
                    names.add(ext_mgd_vl.vnfVirtualLinkDescId)

        return names

    def _make_res_def_for_new_vdu(self, vdu_name, num_inst, cp_names,
            storage_names):
        # common part of instantiate and scale out
        add_reses = []
        for _ in range(num_inst):
            vdu_res_id = uuidutils.generate_uuid()
            add_reses.append(
                objects.ResourceDefinitionV1(
                    id=vdu_res_id,
                    type='COMPUTE',
                    resourceTemplateId=vdu_name
                )
            )
            for cp_name in cp_names:
                add_reses.append(
                    objects.ResourceDefinitionV1(
                        id=_make_combination_id(cp_name, vdu_res_id),
                        type='LINKPORT',
                        resourceTemplateId=cp_name
                    )
                )
            for storage_name in storage_names:
                add_reses.append(
                    objects.ResourceDefinitionV1(
                        id=_make_combination_id(storage_name, vdu_res_id),
                        type='STORAGE',
                        resourceTemplateId=storage_name
                    )
                )

        return add_reses

    def instantiate_grant(self, grant_req, req, inst, vnfd):
        flavour_id = req.flavourId

        if vnfd.get_vnfd_flavour(flavour_id) is None:
            raise sol_ex.FlavourIdNotFound(flavour_id=flavour_id)

        grant_req.flavourId = flavour_id

        if req.obj_attr_is_set('instantiationLevelId'):
            inst_level = req.instantiationLevelId
            grant_req.instantiationLevelId = inst_level
        else:
            inst_level = vnfd.get_default_instantiation_level(flavour_id)

        add_reses = []
        nodes = vnfd.get_vdu_nodes(flavour_id)
        link_port_names = self._get_link_ports(req)
        for name, node in nodes.items():
            num = vnfd.get_vdu_num(flavour_id, name, inst_level)
            vdu_cp_names = vnfd.get_vdu_cps(flavour_id, name)
            vdu_storage_names = vnfd.get_vdu_storages(node)

            add_reses += self._make_res_def_for_new_vdu(name, num,
                set(vdu_cp_names) - link_port_names, vdu_storage_names)

        ext_mgd_vls = []
        if req.obj_attr_is_set('extManagedVirtualLinks'):
            ext_mgd_vls = [ext_mgd_vl.vnfVirtualLinkDescId
                           for ext_mgd_vl in req.extManagedVirtualLinks]
        nodes = vnfd.get_virtual_link_nodes(flavour_id)
        for name in nodes.keys():
            if name in ext_mgd_vls:
                continue
            res_def = objects.ResourceDefinitionV1(
                id=uuidutils.generate_uuid(),
                type='VL',
                resourceTemplateId=name)
            add_reses.append(res_def)

        if add_reses:
            grant_req.addResources = add_reses

        # placementConstraints
        affinity_policies = {
            'AFFINITY': vnfd.get_affinity_targets(flavour_id),
            'ANTI_AFFINITY': vnfd.get_anti_affinity_targets(flavour_id)
        }
        plc_consts = []
        for key, value in affinity_policies.items():
            for targets, scope in value:
                res_refs = []
                for target in targets:
                    for res in add_reses:
                        if res.resourceTemplateId == target:
                            res_ref = objects.ConstraintResourceRefV1(
                                idType='GRANT',
                                resourceId=res.id)
                            res_refs.append(res_ref)

                plc_const = objects.PlacementConstraintV1(
                    affinityOrAntiAffinity=key,
                    scope=scope.upper(),
                    resource=res_refs)
                plc_consts.append(plc_const)

        if plc_consts:
            grant_req.placementConstraints = plc_consts

        if req.obj_attr_is_set('additionalParams'):
            grant_req.additionalParams = req.additionalParams

    def instantiate_post_grant(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        # set inst vimConnectionInfo
        req = lcmocc.operationParams
        vim_infos = {}
        if req.obj_attr_is_set('vimConnectionInfo'):
            vim_infos = req.vimConnectionInfo

        if grant.obj_attr_is_set('vimConnectionInfo'):
            # if NFVO returns vimConnectionInfo use it.
            # As the controller does for req.vimConnectionInfo, if accessInfo
            # or interfaceInfo is not specified, get them from VIM DB.
            # vimId must be in VIM DB.
            res_vim_infos = grant.vimConnectioninfo
            for key, res_vim_info in res_vim_infos.items():
                if not (res_vim_info.obj_attr_is_set('accessInfo') and
                        res_vim_info.obj_attr_is_set('interfaceInfo')):
                    vim_info = vim_utils.get_vim(context, res_vim_info.vimId)
                    res_vim_infos[key] = vim_info

            vim_infos = inst_utils.json_merge_patch(vim_infos, res_vim_infos)

        if not vim_infos:
            # use the default VIM for the project. tacker special.
            vim_info = vim_utils.get_default_vim(context)
            if vim_info:
                vim_infos["default"] = vim_info
            else:
                # must be one vimConnectionInfo at least.
                raise sol_ex.NoVimConnectionInfo()

        inst.vimConnectionInfo = vim_infos

    def instantiate_process(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        for attr in ['vnfConfigurableProperties', 'extensions']:
            self._modify_from_req(inst, req, attr)

        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.instantiate(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

        inst.instantiationState = 'INSTANTIATED'

    def instantiate_rollback(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.instantiate_rollback(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

    def _make_res_def_for_remove_vnfcs(self, inst_info, inst_vnfcs,
            re_create=False):
        # common part of terminate, scale-in and heal.
        # only heal calls with re_create=True.
        rm_reses = []
        add_reses = []
        vnfc_cps = {}
        for inst_vnfc in inst_vnfcs:
            rm_vdu_res_id = uuidutils.generate_uuid()
            rm_reses.append(
                objects.ResourceDefinitionV1(
                    id=rm_vdu_res_id,
                    type='COMPUTE',
                    resourceTemplateId=inst_vnfc.vduId,
                    resource=inst_vnfc.computeResource
                )
            )
            if re_create:
                add_vdu_res_id = uuidutils.generate_uuid()
                add_reses.append(
                    objects.ResourceDefinitionV1(
                        id=add_vdu_res_id,
                        type='COMPUTE',
                        resourceTemplateId=inst_vnfc.vduId,
                    )
                )

            if inst_vnfc.obj_attr_is_set('vnfcCpInfo'):
                for cp_info in inst_vnfc.vnfcCpInfo:
                    if not (cp_info.obj_attr_is_set('vnfExtCpId') or
                            cp_info.obj_attr_is_set('vnfLinkPortId')):
                        # it means extLinkPorts of extVirtualLinks was
                        # specified. so it is not the resource to be
                        # deleted.
                        continue
                    res_def = objects.ResourceDefinitionV1(
                        id=_make_combination_id(cp_info.cpdId, rm_vdu_res_id),
                        resourceTemplateId=cp_info.cpdId,
                        type='LINKPORT')
                    rm_reses.append(res_def)
                    if cp_info.obj_attr_is_set('vnfExtCpId'):
                        vnfc_cps[cp_info.vnfExtCpId] = res_def
                    else:  # vnfLinkPortId
                        vnfc_cps[cp_info.vnfLinkPortId] = res_def

                    if re_create:
                        add_reses.append(
                            objects.ResourceDefinitionV1(
                                id=_make_combination_id(cp_info.cpdId,
                                                        add_vdu_res_id),
                                resourceTemplateId=cp_info.cpdId,
                                type='LINKPORT'
                            )
                        )

            if inst_vnfc.obj_attr_is_set('storageResourceIds'):
                for storage_id in inst_vnfc.storageResourceIds:
                    inst_str = _get_obj_by_id(storage_id,
                        inst_info.virtualStorageResourceInfo)
                    str_name = inst_str.virtualStorageDescId
                    rm_reses.append(
                        objects.ResourceDefinitionV1(
                            id=_make_combination_id(str_name, rm_vdu_res_id),
                            type='STORAGE',
                            resourceTemplateId=str_name,
                            resource=inst_str.storageResource
                        )
                    )
                    if re_create:
                        add_reses.append(
                            objects.ResourceDefinitionV1(
                                id=_make_combination_id(str_name,
                                                        add_vdu_res_id),
                                type='STORAGE',
                                resourceTemplateId=str_name
                            )
                        )

        # fill resourceHandle of ports
        if inst_info.obj_attr_is_set('vnfVirtualLinkResourceInfo'):
            for inst_vl in inst_info.vnfVirtualLinkResourceInfo:
                if inst_vl.obj_attr_is_set('vnfLinkPorts'):
                    for port in inst_vl.vnfLinkPorts:
                        if port.id in vnfc_cps:
                            res_def = vnfc_cps[port.id]
                            res_def.resource = port.resourceHandle

        if inst_info.obj_attr_is_set('extVirtualLinkInfo'):
            for ext_vl in inst_info.extVirtualLinkInfo:
                if ext_vl.obj_attr_is_set('extLinkPorts'):
                    for port in ext_vl.extLinkPorts:
                        if (port.obj_attr_is_set('cpInstanceId') and
                                port.cpInstanceId in vnfc_cps):
                            res_def = vnfc_cps[port.cpInstanceId]
                            res_def.resource = port.resourceHandle

        if inst_info.obj_attr_is_set('extManagedVirtualLinkInfo'):
            for ext_mgd_vl in inst_info.extManagedVirtualLinkInfo:
                if ext_mgd_vl.obj_attr_is_set('vnfLinkPorts'):
                    for port in ext_mgd_vl.vnfLinkPorts:
                        if (port.obj_attr_is_set('cpInstanceId') and
                                port.id in vnfc_cps):
                            res_def = vnfc_cps[port.id]
                            res_def.resource = port.resourceHandle

        return rm_reses, add_reses

    def terminate_grant(self, grant_req, req, inst, vnfd):
        inst_info = inst.instantiatedVnfInfo
        rm_reses = []
        if inst_info.obj_attr_is_set('vnfcResourceInfo'):
            rm_reses, _ = self._make_res_def_for_remove_vnfcs(
                inst_info, inst_info.vnfcResourceInfo)

        if inst_info.obj_attr_is_set('vnfVirtualLinkResourceInfo'):
            rm_reses += [
                objects.ResourceDefinitionV1(
                    id=uuidutils.generate_uuid(),
                    type='VL',
                    resourceTemplateId=inst_vl.vnfVirtualLinkDescId,
                    resource=inst_vl.networkResource
                )
                for inst_vl in inst_info.vnfVirtualLinkResourceInfo
            ]

        if rm_reses:
            grant_req.removeResources = rm_reses

    def terminate_process(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.terminate(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

        inst.instantiationState = 'NOT_INSTANTIATED'

        # reset instantiatedVnfInfo
        # NOTE: reset after update lcmocc
        inst_vnf_info = objects.VnfInstanceV2_InstantiatedVnfInfo(
            flavourId=inst.instantiatedVnfInfo.flavourId,
            vnfState='STOPPED',
            # NOTE: extCpInfo is omitted. its cardinality is 1..N but it is
            # meaningless to have it for terminated vnf instance.
        )
        inst.instantiatedVnfInfo = inst_vnf_info

        # reset vimConnectionInfo
        inst.vimConnectionInfo = {}

    def scale_grant(self, grant_req, req, inst, vnfd):
        flavour_id = inst.instantiatedVnfInfo.flavourId
        scale_type = req.type
        aspect_id = req.aspectId
        num_steps = req.numberOfSteps

        vdu_num_inst = vnfd.get_scale_vdu_and_num(flavour_id, aspect_id)
        if not vdu_num_inst:
            # should not occur. just check for consistency.
            raise sol_ex.InvalidScaleAspectId(aspect_id=aspect_id)

        if scale_type == 'SCALE_OUT':
            self._make_scale_out_grant_request(grant_req, inst, num_steps,
                vdu_num_inst)
        else:
            self._make_scale_in_grant_request(grant_req, inst, num_steps,
                vdu_num_inst)

        if req.obj_attr_is_set('additionalParams'):
            grant_req.additionalParams = req.additionalParams

    def _make_scale_out_grant_request(self, grant_req, inst, num_steps,
            vdu_num_inst):
        inst_info = inst.instantiatedVnfInfo
        add_reses = []

        # get one of vnfc for the vdu from inst.instantiatedVnfInfo
        vdu_sample = {}
        for vdu_name in vdu_num_inst.keys():
            for inst_vnfc in inst_info.vnfcResourceInfo:
                if inst_vnfc.vduId == vdu_name:
                    vdu_sample[vdu_name] = inst_vnfc
                    break

        for vdu_name, inst_vnfc in vdu_sample.items():
            num_inst = vdu_num_inst[vdu_name] * num_steps

            vdu_cp_names = []
            if inst_vnfc.obj_attr_is_set('vnfcCpInfo'):
                # NOTE: it is expected that there are only dynamic ports
                # for vdus which enable scaling.
                vdu_cp_names = [cp_info.cpdId
                                for cp_info in inst_vnfc.vnfcCpInfo]

            vdu_storage_names = []
            if inst_vnfc.obj_attr_is_set('storageResourceIds'):
                for storage_id in inst_vnfc.storageResourceIds:
                    storage_res = _get_obj_by_id(storage_id,
                        inst_info.virtualStorageResourceInfo)
                    vdu_storage_names.append(
                        storage_res.virtualStorageDescId)

            add_reses += self._make_res_def_for_new_vdu(vdu_name,
                    num_inst, vdu_cp_names, vdu_storage_names)

        if add_reses:
            grant_req.addResources = add_reses

    def _make_scale_in_grant_request(self, grant_req, inst, num_steps,
            vdu_num_inst):
        inst_info = inst.instantiatedVnfInfo
        rm_vnfcs = []

        # select remove VDUs
        # NOTE: scale-in specification of tacker SOL003 v2 API is that
        # newer VDU is selected for reduction.
        # It is expected that vnfcResourceInfo is sorted by creation_time
        # of VDU, newer is earlier.
        for vdu_name, num_inst in vdu_num_inst.items():
            num_inst = num_inst * num_steps

            count = 0
            for inst_vnfc in inst_info.vnfcResourceInfo:
                if inst_vnfc.vduId == vdu_name:
                    rm_vnfcs.append(inst_vnfc)
                    count += 1
                    if count == num_inst:
                        break

        rm_reses, _ = self._make_res_def_for_remove_vnfcs(inst_info, rm_vnfcs)

        if rm_reses:
            grant_req.removeResources = rm_reses

    def scale_process(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.scale(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

    def scale_rollback(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        if req.type == 'SCALE_IN':
            raise sol_ex.RollbackNotSupported(op='SCALE IN')

        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.scale_rollback(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

    def _modify_from_vnfd_prop(self, inst, vnfd_prop, attr):
        if vnfd_prop.get(attr):
            setattr(inst, attr, vnfd_prop[attr])
        elif inst.obj_attr_is_set(attr):
            # set {} since attribute deletion is not supported.
            setattr(inst, attr, {})

    def _modify_from_req(self, inst, req, attr):
        if req.obj_attr_is_set(attr):
            base = getattr(inst, attr) if inst.obj_attr_is_set(attr) else {}
            setattr(inst, attr, inst_utils.json_merge_patch(
                base, getattr(req, attr)))

    def _merge_vim_connection_info(self, inst, req):
        # used by MODIFY_INFO and CHANGE_EXT_CONN
        # inst.vimConnectionInfo exists since req.vimConnectionInfo
        # can be set only if vnf instance is INSTANTIATED.
        inst_viminfo = inst.to_dict()['vimConnectionInfo']
        req_viminfo = req.to_dict()['vimConnectionInfo']
        merge = inst_utils.json_merge_patch(inst_viminfo, req_viminfo)
        inst.vimConnectionInfo = {
            key: objects.VimConnectionInfo.from_dict(value)
            for key, value in merge.items()}

    def modify_info_process(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams

        if req.obj_attr_is_set('vnfInstanceName'):
            inst.vnfInstanceName = req.vnfInstanceName

        if req.obj_attr_is_set('vnfInstanceDescription'):
            inst.vnfInstanceDescription = req.vnfInstanceDescription

        if req.obj_attr_is_set('vnfdId') and req.vnfdId != inst.vnfdId:
            # NOTE: When vnfdId is changed, the values of attributes
            # in the VnfInstance needs update from new VNFD.
            inst.vnfdId = req.vnfdId

            pkg_info = self.nfvo_client.get_vnf_package_info_vnfd(
                context, inst.vnfdId)

            new_vnfd = self.nfvo_client.get_vnfd(context, inst.vnfdId)
            new_vnfd_prop = new_vnfd.get_vnfd_properties()

            inst.vnfProvider = pkg_info.vnfProvider
            inst.vnfProductName = pkg_info.vnfProductName
            inst.vnfSoftwareVersion = pkg_info.vnfSoftwareVersion
            inst.vnfdVersion = pkg_info.vnfdVersion

            attrs = ['vnfConfigurableProperties', 'metadata', 'extensions']
            for attr in attrs:
                self._modify_from_vnfd_prop(inst, new_vnfd_prop, attr)

        attrs = ['vnfConfigurableProperties', 'metadata', 'extensions']
        for attr in attrs:
            self._modify_from_req(inst, req, attr)

        if req.obj_attr_is_set('vimConnectionInfo'):
            self._merge_vim_connection_info(inst, req)

        if req.obj_attr_is_set('vnfcInfoModifications'):
            for vnfc_mod in req.vnfcInfoModifications:
                # existence of ids are checked by controller
                for vnfc_info in inst.instantiatedVnfInfo.vnfcInfo:
                    if vnfc_mod.id == vnfc_info.id:
                        prop_base = {}
                        if vnfc_info.obj_attr_is_set(
                                'vnfcConfigurableProperties'):
                            prop_base = vnfc_info.vnfcConfigurableProperties
                        vnfc_info.vnfcConfigurableProperties = (
                            inst_utils.json_merge_patch(prop_base,
                                vnfc_mod.vnfcConfigurableProperties))
                        break

    def modify_info_rollback(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        # DB is not updated, rollback does nothing and makes it successful.
        pass

    def _set_rm_add_reses(self, rm_reses, add_reses, res_type, res_name,
            res_obj, rm_id, add_id):
        rm_reses.append(
            objects.ResourceDefinitionV1(
                id=rm_id,
                type=res_type,
                resourceTemplateId=res_name,
                resource=res_obj
            )
        )
        add_reses.append(
            objects.ResourceDefinitionV1(
                id=add_id,
                type=res_type,
                resourceTemplateId=res_name
            )
        )

    def _make_SOL002_heal_grant_request(self, grant_req, req, inst_info,
            is_all):
        rm_reses = []
        add_reses = []

        for vnfc in inst_info.vnfcInfo:
            if vnfc.id not in req.vnfcInstanceId:
                continue
            inst_vnfc = _get_obj_by_id(vnfc.vnfcResourceInfoId,
                                       inst_info.vnfcResourceInfo)
            rm_vdu_res_id = uuidutils.generate_uuid()
            add_vdu_res_id = uuidutils.generate_uuid()
            self._set_rm_add_reses(rm_reses, add_reses, 'COMPUTE',
                inst_vnfc.vduId, inst_vnfc.computeResource,
                rm_vdu_res_id, add_vdu_res_id)

            if is_all and inst_vnfc.obj_attr_is_set('storageResourceIds'):
                for storage_id in inst_vnfc.storageResourceIds:
                    inst_str = _get_obj_by_id(storage_id,
                        inst_info.virtualStorageResourceInfo)
                    str_name = inst_str.virtualStorageDescId
                    self._set_rm_add_reses(rm_reses, add_reses, 'STORAGE',
                        str_name, inst_str.storageResource,
                        _make_combination_id(str_name, rm_vdu_res_id),
                        _make_combination_id(str_name, add_vdu_res_id))

        if rm_reses:
            grant_req.removeResources = rm_reses
            grant_req.addResources = add_reses

    def _make_SOL003_heal_grant_request(self, grant_req, req, inst_info,
            is_all):
        rm_reses = []
        add_reses = []

        if is_all:
            if inst_info.obj_attr_is_set('vnfcResourceInfo'):
                rm_reses, add_reses = self._make_res_def_for_remove_vnfcs(
                    inst_info, inst_info.vnfcResourceInfo, re_create=True)

            if inst_info.obj_attr_is_set('vnfVirtualLinkResourceInfo'):
                for inst_vl in inst_info.vnfVirtualLinkResourceInfo:
                    self._set_rm_add_reses(rm_reses, add_reses, 'VL',
                        inst_vl.vnfVirtualLinkDescId, inst_vl.networkResource,
                        uuidutils.generate_uuid(), uuidutils.generate_uuid())
        else:
            if inst_info.obj_attr_is_set('vnfcResourceInfo'):
                for inst_vnfc in inst_info.vnfcResourceInfo:
                    self._set_rm_add_reses(rm_reses, add_reses, 'COMPUTE',
                        inst_vnfc.vduId, inst_vnfc.computeResource,
                        uuidutils.generate_uuid(), uuidutils.generate_uuid())

        if rm_reses:
            grant_req.removeResources = rm_reses
            grant_req.addResources = add_reses

    def heal_grant(self, grant_req, req, inst, vnfd):
        inst_info = inst.instantiatedVnfInfo

        is_all = False  # default is False
        if req.obj_attr_is_set('additionalParams'):
            is_all = req.additionalParams.get('all', False)
            grant_req.additionalParams = req.additionalParams

        if req.obj_attr_is_set('vnfcInstanceId'):
            self._make_SOL002_heal_grant_request(grant_req, req, inst_info,
                                                 is_all)
        else:
            self._make_SOL003_heal_grant_request(grant_req, req, inst_info,
                                                 is_all)

    def heal_process(self, context, lcmocc, inst, grant_req, grant, vnfd):
        req = lcmocc.operationParams
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.heal(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

    def change_ext_conn_grant(self, grant_req, req, inst, vnfd):
        inst_info = inst.instantiatedVnfInfo

        cp_names = set()
        link_ports = set()
        for ext_vl in req.extVirtualLinks:
            for ext_cp in ext_vl.extCps:
                cp_names.add(ext_cp.cpdId)
                for cp_config in ext_cp.cpConfig.values():
                    if cp_config.obj_attr_is_set('linkPortId'):
                        link_ports.add(ext_cp.cpdId)

        add_reses = []
        rm_reses = []
        update_reses = []
        rm_vnfc_cps = {}
        for inst_vnfc in inst_info.vnfcResourceInfo:
            if not inst_vnfc.obj_attr_is_set('vnfcCpInfo'):
                continue
            vnfc_cps = {cp_info.cpdId for cp_info in inst_vnfc.vnfcCpInfo}
            if not vnfc_cps & cp_names:
                continue

            old_vdu_res_id = uuidutils.generate_uuid()
            new_vdu_res_id = uuidutils.generate_uuid()
            update_reses.append(
                objects.ResourceDefinitionV1(
                    id=new_vdu_res_id,
                    type='COMPUTE',
                    resourceTemplateId=inst_vnfc.vduId,
                    resource=inst_vnfc.computeResource
                )
            )

            for cp_info in inst_vnfc.vnfcCpInfo:
                if cp_info.cpdId not in cp_names:
                    continue

                if cp_info.obj_attr_is_set('vnfExtCpId'):
                    # if there is not vnfExtCpId, it means extLinkPorts of
                    # extVirtualLinks was specified.
                    res_def = objects.ResourceDefinitionV1(
                        id=_make_combination_id(cp_info.cpdId, old_vdu_res_id),
                        resourceTemplateId=cp_info.cpdId,
                        type='LINKPORT')
                    rm_reses.append(res_def)
                    rm_vnfc_cps[cp_info.vnfExtCpId] = res_def

                if cp_info.cpdId not in link_ports:
                    add_reses.append(
                        objects.ResourceDefinitionV1(
                            id=_make_combination_id(cp_info.cpdId,
                                                    new_vdu_res_id),
                            resourceTemplateId=cp_info.cpdId,
                            type='LINKPORT'
                        )
                    )

        # fill resourceHandle of rm_reses
        if inst_info.obj_attr_is_set('extVirtualLinkInfo'):
            for ext_vl in inst_info.extVirtualLinkInfo:
                if ext_vl.obj_attr_is_set('extLinkPorts'):
                    for port in ext_vl.extLinkPorts:
                        if (port.obj_attr_is_set('cpInstanceId') and
                                port.cpInstanceId in rm_vnfc_cps):
                            res_def = rm_vnfc_cps[port.cpInstanceId]
                            res_def.resource = port.resourceHandle

        if add_reses:
            grant_req.addResources = add_reses
        if rm_reses:
            grant_req.removeResources = rm_reses
        if update_reses:
            grant_req.updateResources = update_reses

        if req.obj_attr_is_set('additionalParams'):
            grant_req.additionalParams = req.additionalParams

    def change_ext_conn_process(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        if req.obj_attr_is_set('vimConnectionInfo'):
            self._merge_vim_connection_info(inst, req)

        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.change_ext_conn(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')

    def change_ext_conn_rollback(self, context, lcmocc, inst, grant_req,
            grant, vnfd):
        req = lcmocc.operationParams
        vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
        if vim_info.vimType == 'ETSINFV.OPENSTACK_KEYSTONE.V_3':
            driver = openstack.Openstack()
            driver.change_ext_conn_rollback(req, inst, grant_req, grant, vnfd)
        else:
            # only support openstack at the moment
            raise sol_ex.SolException(sol_detail='not support vim type')



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from tacker.api.validation import parameter_types


# SOL013 7.2.2
Identifier = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

# SOL003 4.4.2.2
IdentifierInVnfd = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

# SOL003 4.4.2.2
IdentifierInVim = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

# SOL003 4.4.2.2
IdentifierInVnf = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

# SOL003 4.4.2.2
IdentifierLocal = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

# SOL003 4.4.1.7
ResourceHandle = {
    'type': 'object',
    'properties': {
        'vimConnectionId': Identifier,
        'resourceProviderId': Identifier,
        'resourceId': IdentifierInVim,
        'vimLevelResourceType': {'type': 'string', 'maxLength': 255},
    },
    'required': ['resourceId'],
    'additionalProperties': True,
}

# SOL003 4.4.1.6
VimConnectionInfo = {
    'type': 'object',
    'properties': {
        'vimId': {'type': 'string', 'maxLength': 255},
        'vimType': {'type': 'string', 'minLength': 1, 'maxLength': 255},
        'interfaceInfo': parameter_types.keyvalue_pairs,
        'accessInfo': parameter_types.keyvalue_pairs,
        'extra': parameter_types.keyvalue_pairs,
    },
    'required': ['vimType'],
    'additionalProperties': True,
}

# SOL003 4.4.1.10c (inner)
_IpAddresses = {
    'type': 'object',
    'properties': {
        'type': {'enum': ('IPV4', 'IPV6')},
        'fixedAddresses': {'type': 'array'},
        'numDynamicAddresses': parameter_types.positive_integer,
        'addressRange': {'type': 'object'},
        'subnetId': IdentifierInVim
    },
    'if': {'properties': {'type': {'const': 'IPV4'}}},
    'then': {
        'properties': {
            'fixedAddresses': {
                'type': 'array',
                'items': {'type': 'string', 'format': 'ipv4'}
            },
            'addressRange': {
                'type': 'object',
                'properties': {
                    'minAddress': {'type': 'string', 'format': 'ipv4'},
                    'maxAddress': {'type': 'string', 'format': 'ipv4'}
                },
                'required': ['minAddress', 'maxAddress'],
                'additionalProperties': True
            },
        }
    },
    'else': {
        'properties': {
            'fixedAddresses': {
                'type': 'array',
                'items': {'type': 'string', 'format': 'ipv6'}
            },
            'addressRange': {
                'type': 'object',
                'properties': {
                    'minAddress': {'type': 'string', 'format': 'ipv6'},
                    'maxAddress': {'type': 'string', 'format': 'ipv6'}
                },
                'required': ['minAddress', 'maxAddress'],
                'additionalProperties': True
            },
        }
    },
    'required': ['type'],
    'oneOf': [
        {'required': ['numDynamicAddresses']},
        {'required': ['fixedAddresses']},
        {'required': ['addressRange']},
    ],
    'additionalProperties': True
}

# SOL003 4.4.1.10c
IpOverEthernetAddressData = {
    'type': 'object',
    'properties': {
        'macAddress': {'type': 'string', 'format': 'mac_address'},
        'segmentationId': {'type': 'string'},
        'ipAddresses': {
            'type': 'array',
            'items': _IpAddresses}
    },
    'anyOf': [
        {'required': ['macAddress']},
        {'required': ['ipAddresses']}
    ],
    'additionalProperties': True
}

# SOL003 4.4.1.10b
CpProtocolData = {
    'type': 'object',
    'properties': {
        'layerProtocol': {
            'type': 'string',
            'enum': 'IP_OVER_ETHERNET'},
        'ipOverEthernet': IpOverEthernetAddressData,
    },
    'required': ['layerProtocol'],
    'additionalProperties': True,
}

# SOL003 4.4.1.10a
VnfExtCpConfig = {
    'type': 'object',
    'properties': {
        'parentCpConfigId': IdentifierInVnf,
        'linkPortId': Identifier,
        'cpProtocolData': {
            'type': 'array',
            'items': CpProtocolData}
    },
    'additionalProperties': True
}

# SOL003 4.4.1.10
VnfExtCpData = {
    'type': 'object',
    'properties': {
        'cpdId': IdentifierInVnfd,
        'cpConfig': {
            'type': 'object',
            'minProperties': 1,
            'patternProperties': {
                '^.*$': VnfExtCpConfig
            }
        }
    },
    'required': ['cpdId', 'cpConfig'],
    'additionalProperties': True
}

# SOL003 4.4.1.14
ExtLinkPortData = {
    'type': 'object',
    'properties': {
        'id': Identifier,
        'resourceHandle': ResourceHandle,
    },
    'required': ['id', 'resourceHandle'],
    'additionalProperties': True,
}

# SOL003 4.4.1.11
ExtVirtualLinkData = {
    'type': 'object',
    'properties': {
        'id': Identifier,
        'vimConnectionId': Identifier,
        'resourceProviderId': Identifier,
        'resourceId': IdentifierInVim,
        'extCps': {
            'type': 'array',
            'minItems': 1,
            'items': VnfExtCpData},
        'extLinkPorts': {
            'type': 'array',
            'items': ExtLinkPortData}
    },
    'required': ['id', 'resourceId', 'extCps'],
    'additionalProperties': True
}

# SOL003 5.5.3.18
VnfLinkPortData = {
    'type': 'object',
    'properties': {
        'vnfLinkPortId': Identifier,
        'resourceHandle': ResourceHandle
    },
    'required': ['vnfLinkPortId', 'resourceHandle'],
    'additionalProperties': True,
}

# SOL003 4.4.1.12
ExtManagedVirtualLinkData = {
    'type': 'object',
    'properties': {
        'id': Identifier,
        'vnfVirtualLinkDescId': IdentifierInVnfd,
        'vimConnectionId': Identifier,
        'resourceProviderId': Identifier,
        'resourceId': IdentifierInVim,
        'vnfLinkPort': {
            'type': 'array',
            'items': VnfLinkPortData},
        'extManagedMultisiteVirtualLinkId': Identifier
    },
    'required': ['id', 'vnfVirtualLinkDescId', 'resourceId'],
    'additionalProperties': True,
}

# SOL002 5.5.3.24
VnfcInfoModifications = {
    'type': 'object',
    'properties': {
        'id': IdentifierInVnf,
        'vnfcConfigurableProperties': parameter_types.keyvalue_pairs,
    },
    'required': ['id', 'vnfcConfigurableProperties'],
    'additionalProperties': True,
}



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
The module provides all database models at current HEAD.

Its purpose is to create comparable metadata with current database schema.
Based on this comparison database can be healed with healing migration.

"""

from tacker.db import model_base  # noqa
from tacker.db.nfvo import nfvo_db  # noqa
from tacker.db.nfvo import ns_db  # noqa
from tacker.db.nfvo import vnffg_db  # noqa
from tacker.db.vnfm import vnfm_db  # noqa
from tacker.sol_refactored.db.sqlalchemy import models  # noqa



# Copyright (c) 2020 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestInstantiatedVnfInfo(SqlTestCase):

    def setUp(self):
        super(TestInstantiatedVnfInfo, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_package = self._create_and_upload_vnf_package()
        self.vnf_instance = self._create_vnf_instance()
        self.resource_handle_info = self._create_resource_handle()
        self.ext_link_port_info = self._create_ext_link_port_info()
        self.ext_virtual_link_info = self._create_ext_virtual_link_info()
        self.vnf_link_ports_info = self._create_vnf_link_ports()
        self.ip_addresses_info = self._create_ip_addresses_info()
        self.ip_over_ethernet = self._create_ip_over_ethernet_info()
        self.cp_protocol_info = self._create_cp_protocol_info()
        self.vnf_external_cp_info = self._create_vnf_external_cp_info()
        self.vnfc_cp_info = self._create_vnfc_cp_info()
        self.vnfc_resource_info = self._create_vnfc_resource_info()
        self.virtual_link_resource_info = \
            self._create_virtual_link_resource_info()
        self.virtual_storage_resource_info = \
            self._create_virtual_storage_resource_info()
        self.ext_managed_virtual_link_info = \
            self._create_ext_managed_virtual_link_info()

    def _create_and_upload_vnf_package(self):
        vnf_package = objects.VnfPackage(context=self.context,
                                         **fakes.vnf_package_data)
        vnf_package.create()

        vnf_pack_vnfd = fakes.get_vnf_package_vnfd_data(
            vnf_package.id, uuidsentinel.vnfd_id)

        vnf_pack_vnfd_obj = objects.VnfPackageVnfd(
            context=self.context, **vnf_pack_vnfd)
        vnf_pack_vnfd_obj.create()
        self.vnf_package_vnfd = vnf_pack_vnfd_obj
        vnf_package.vnf_package = "ONBOARDED"
        vnf_package.save()

        return vnf_package

    def _create_vnf_instance(self):
        vnf_instance_data = fakes.get_vnf_instance_data(
            self.vnf_package_vnfd.vnfd_id)
        vnf_instance = objects.VnfInstance(context=self.context,
                                           **vnf_instance_data)
        vnf_instance.create()
        return vnf_instance

    def _create_vnf_external_cp_info(self):
        vnf_external_cp_data = copy.deepcopy(fakes.vnf_external_cp_info)
        vnf_external_cp_data.update(
            {'cp_protocol_info': [self.cp_protocol_info]})
        vnf_external_cp_info = objects.VnfExtCpInfo(
            context=self.context, **vnf_external_cp_data)
        return vnf_external_cp_info

    def _create_resource_handle(self):
        resource_handle_data = copy.deepcopy(fakes.resource_handle_info)
        resource_handle_info = objects.ResourceHandle(
            context=self.context, **resource_handle_data)
        return resource_handle_info

    def _create_ext_link_port_info(self):
        ext_link_port_info = copy.deepcopy(fakes.ext_link_port_info)
        ext_link_port_info.update(
            {'resource_handle': self.resource_handle_info})
        ext_link_port_info = objects.ExtLinkPortInfo(
            context=self.context, **ext_link_port_info)
        return ext_link_port_info

    def _create_ext_virtual_link_info(self):
        ext_virtual_link_info = copy.deepcopy(fakes.ext_virtual_link_info)
        ext_virtual_link_info.update(
            {'resource_handle_info': self.resource_handle_info,
             'ext_link_ports': self.ext_link_port_info})
        ext_virtual_link_info = objects.VnfExtCpInfo(
            context=self.context, **ext_virtual_link_info)
        return ext_virtual_link_info

    def _create_vnf_link_ports(self):
        vnf_link_ports_info = copy.deepcopy(fakes.vnf_link_ports)
        vnf_link_ports_info.update(
            {'resource_handle': self.resource_handle_info})
        vnf_link_ports_info = objects.VnfLinkPortInfo(
            context=self.context, **vnf_link_ports_info)
        return vnf_link_ports_info

    def _create_ext_managed_virtual_link_info(self):
        ext_managed_virtual_link_info = copy.deepcopy(
            fakes.ext_managed_virtual_link_info)
        ext_managed_virtual_link_info.update(
            {'network_resource': self.resource_handle_info,
             'vnf_link_ports': [self.vnf_link_ports_info]})
        ext_managed_virtual_link_info = objects.ExtManagedVirtualLinkInfo(
            context=self.context, **ext_managed_virtual_link_info)
        return ext_managed_virtual_link_info

    def _create_ip_addresses_info(self):
        ip_address_info = copy.deepcopy(fakes.ip_address_info)
        ip_address_info = objects.IpAddress(
            context=self.context, **ip_address_info)
        return ip_address_info

    def _create_ip_over_ethernet_info(self):
        ip_over_ethernet_onfo = copy.deepcopy(
            fakes.ip_over_ethernet_address_info)
        ip_over_ethernet_onfo.update(
            {'ip_addresses': [self.ip_addresses_info]})
        ip_over_ethernet_onfo = objects.IpOverEthernetAddressInfo(
            context=self.context, **ip_over_ethernet_onfo)
        return ip_over_ethernet_onfo

    def _create_cp_protocol_info(self):
        cp_protocol_info = copy.deepcopy(fakes.cp_protocol_info)
        cp_protocol_info.update(
            {'ip_over_ethernet': self.ip_over_ethernet})
        cp_protocol_info = objects.CpProtocolInfo(
            context=self.context, **cp_protocol_info)
        return cp_protocol_info

    def _create_vnfc_cp_info(self):
        vnfc_cp_info = copy.deepcopy(fakes.vnfc_cp_info)
        vnfc_cp_info.update(
            {'cp_protocol_info': [self.cp_protocol_info]})
        vnfc_cp_info = objects.VnfcCpInfo(
            context=self.context, **vnfc_cp_info)
        return vnfc_cp_info

    def _create_vnfc_resource_info(self):
        vnfc_resource_info = copy.deepcopy(fakes.vnfc_resource_info)
        vnfc_resource_info.update(
            {'compute_resource': self.resource_handle_info,
             'vnf_link_ports': [self.vnf_link_ports_info],
             'vnfc_cp_info': [self.vnfc_cp_info]})
        vnfc_resource_info = objects.VnfcResourceInfo(
            context=self.context, **vnfc_resource_info)
        return vnfc_resource_info

    def _create_virtual_link_resource_info(self):
        vnf_virtual_link_resource_info = copy.deepcopy(
            fakes.vnf_virtual_link_resource_info)
        vnf_virtual_link_resource_info.update(
            {'network_resource': self.resource_handle_info,
             'vnf_link_ports': [self.vnf_link_ports_info]})
        vnf_virtual_link_resource_info = objects.VnfVirtualLinkResourceInfo(
            context=self.context, **vnf_virtual_link_resource_info)
        return vnf_virtual_link_resource_info

    def _create_virtual_storage_resource_info(self):
        virtual_storage_resource_info = copy.deepcopy(
            fakes.virtual_storage_resource_info)
        virtual_storage_resource_info.update(
            {'storage_resource': self.resource_handle_info})
        virtual_storage_resource_info = objects.VirtualStorageResourceInfo(
            context=self.context, **virtual_storage_resource_info)
        return virtual_storage_resource_info

    def test_save(self):
        instantiated_vnf_info = copy.deepcopy(
            fakes.get_instantiated_vnf_info())
        instantiated_vnf_info.update(
            {'ext_cp_info': [self.vnf_external_cp_info],
             'vnf_instance_id': self.vnf_instance.id,
             'ext_link_port_info': self.ext_link_port_info,
             'ext_managed_virtual_link_info': [
                 self.ext_managed_virtual_link_info],
             'vnfc_resource_info': [self.vnfc_resource_info],
             'vnf_virtual_link_resource_info': [
                 self.virtual_link_resource_info],
             'virtual_storage_resource_info': [
                 self.virtual_storage_resource_info]})
        instantiated_vnf_info = objects.InstantiatedVnfInfo(
            context=self.context, **instantiated_vnf_info)
        instantiated_vnf_info.save()
        self.assertIsNotNone(instantiated_vnf_info.created_at)

    def test_resource_handle_obj_from_primitive_and_object_to_dict(self):
        resource_handle = copy.deepcopy(fakes.resource_handle_info)
        result = objects.ResourceHandle.obj_from_primitive(
            resource_handle, self.context)
        self.assertIsInstance(result, objects.ResourceHandle)
        self.assertEqual('TEST', result.vim_level_resource_type)
        resource_handle_dict = result.to_dict()
        self.assertIsInstance(resource_handle_dict, dict)
        self.assertEqual(
            'TEST', resource_handle_dict['vim_level_resource_type'])

    def test_virt_strg_res_info_obj_from_primitive_and_obj_to_dict(self):
        virtual_storage_resource_info = copy.deepcopy(
            fakes.virtual_storage_resource_info)
        result = objects.VirtualStorageResourceInfo.obj_from_primitive(
            virtual_storage_resource_info, self.context)
        self.assertIsInstance(result,
            objects.VirtualStorageResourceInfo)
        virt_strg_res_info_dict = result.to_dict()
        self.assertIsInstance(virt_strg_res_info_dict, dict)

    def test_vnfc_cp_info_obj_from_primitive_and_obj_to_dict(self):
        vnfc_cp_info = copy.deepcopy(fakes.vnfc_cp_info)
        result = objects.VnfcCpInfo.obj_from_primitive(
            vnfc_cp_info, self.context)
        self.assertIsInstance(result, objects.VnfcCpInfo)
        vnfc_cp_info = result.to_dict()
        self.assertIsInstance(vnfc_cp_info, dict)

    def test_vnfc_resource_info_obj_from_primitive_and_obj_to_dict(self):
        vnfc_resource_info = copy.deepcopy(fakes.vnfc_resource_info)
        result = objects.VnfcResourceInfo.obj_from_primitive(
            vnfc_resource_info, self.context)
        self.assertIsInstance(result, objects.VnfcResourceInfo)
        self.assertEqual({'key': 'value'}, result.metadata)
        vnfc_resource_info = result.to_dict()
        self.assertIsInstance(vnfc_resource_info, dict)

    def test_ext_mng_virt_link_obj_from_primitive_and_obj_to_dict(self):
        ext_managed_virtual_link_info = copy.deepcopy(
            fakes.ext_managed_virtual_link_info)
        result = objects.ExtManagedVirtualLinkInfo.obj_from_primitive(
            ext_managed_virtual_link_info, self.context)
        self.assertIsInstance(result, objects.ExtManagedVirtualLinkInfo)
        ext_mng_virt_link = result.to_dict()
        self.assertIsInstance(ext_mng_virt_link, dict)

    def test_ext_link_port_info_obj_from_primitive_and_obj_to_dict(self):
        ext_link_port_info_data = copy.deepcopy(fakes.ext_link_port_info)
        result = objects.ExtLinkPortInfo.obj_from_primitive(
            ext_link_port_info_data, self.context)
        self.assertIsInstance(result, objects.ExtLinkPortInfo)
        ext_link_port_info = result.to_dict()
        self.assertIsInstance(ext_link_port_info, dict)

    def test_ext_virt_link_info_obj_from_primitive_and_obj_to_dict(self):
        ext_virtual_link_info = copy.deepcopy(fakes.ext_virtual_link_info)
        result = objects.ExtVirtualLinkInfo.obj_from_primitive(
            ext_virtual_link_info, self.context)
        self.assertIsInstance(result, objects.ExtVirtualLinkInfo)
        ext_virt_link_info = result.to_dict()
        self.assertIsInstance(ext_virt_link_info, dict)

    def test_vnf_ext_cp_info_obj_from_primitive_and_obj_to_dict(self):
        vnf_ext_cp_info = copy.deepcopy(fakes.vnf_ext_cp_info)
        result = objects.VnfExtCpInfo.obj_from_primitive(
            vnf_ext_cp_info, self.context)
        self.assertIsInstance(result, objects.VnfExtCpInfo)
        ext_virt_link_info = result.to_dict()
        self.assertIsInstance(ext_virt_link_info, dict)

    def test_instantiated_info_obj_from_primitive_and_obj_to_dict(self):
        instantiated_vnf_info = copy.deepcopy(fakes.instantiated_vnf_info)
        result = objects.InstantiatedVnfInfo.obj_from_primitive(
            instantiated_vnf_info, self.context)
        self.assertIsInstance(result, objects.InstantiatedVnfInfo)
        instantiated_vnf_info_dict = result.to_dict()
        self.assertIsInstance(instantiated_vnf_info_dict, dict)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils

from oslo_log import log as logging

from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.db.vnfm import vnfm_db
from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


@db_api.context_manager.writer
def _vnfd_create(context, values):
    vnfd = vnfm_db.VNFD()

    vnfd.update(values)
    vnfd.save(context.session)

    return vnfd


@db_api.context_manager.reader
def _get_vnfd_id(context, id):
    try:
        vnf_package_vnfd = \
            api.model_query(context, models.VnfPackageVnfd).\
            filter_by(package_uuid=id).first()
    except Exception:
        LOG.info("select vnf_package_vnfd failed")
    if vnf_package_vnfd:
        return vnf_package_vnfd.vnfd_id
    else:
        return None


@db_api.context_manager.reader
def _check_vnfd(context, id):
    try:
        vnfd = api.model_query(context, vnfm_db.VNFD).filter_by(id=id).first()
    except Exception:
        LOG.info("select vnfd failed")
    if vnfd:
        return "TRUE"
    else:
        return "FALSE"


@db_api.context_manager.writer
def _vnfd_delete(context, id):
    try:
        api.model_query(context, vnfm_db.VNFD).filter_by(id=id).delete()
    except Exception:
        LOG.info("delete vnfd failed")


@db_api.context_manager.writer
def _vnfd_destroy(context, id):
    now = timeutils.utcnow()
    updated_values = {'deleted_at': now}
    try:
        api.model_query(context, vnfm_db.VNFD).\
            filter_by(id=id).\
            update(updated_values, synchronize_session=False)
    except Exception:
        LOG.info("destroy vnfdfailed")


@base.TackerObjectRegistry.register
class Vnfd(base.TackerObject, base.TackerObjectDictCompat,
        base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'tenant_id': fields.UUIDField(nullable=False),
        'name': fields.StringField(nullable=False),
        'description': fields.StringField(nullable=True),
        'mgmt_driver': fields.StringField(nullable=True),
        'deleted_at': fields.DateTimeField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, vnfd, db_vnfd):

        for key in vnfd.fields:
            if db_vnfd.get(key):
                setattr(vnfd, key, db_vnfd[key])

        vnfd._context = context
        vnfd.obj_reset_changes()

        return vnfd

    @base.remotable
    def create(self):
        updates = self.obj_get_changes()
        db_vnfd = _vnfd_create(
            self._context, updates)
        self._from_db_object(self._context, self, db_vnfd)

    @classmethod
    def obj_from_db_obj(cls, context, db_obj):
        return cls._from_db_object(context, cls(), db_obj)

    @base.remotable
    def destroy(self, id):
        _vnfd_destroy(self._context, id)

    @base.remotable
    def delete(self, id):
        _vnfd_delete(self._context, id)

    @base.remotable
    def check_vnfd(self, id):
        return _check_vnfd(self._context, id)



#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import time
import unittest

from oslo_serialization import jsonutils

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base


class VnfTestAlarmMonitor(base.BaseTackerTest):

    def _test_vnf_tosca_alarm(self, vnfd_file, vnf_name):
        vnfd_instance, vnf_instance, tosca_dict = self.vnfd_and_vnf_create(
            vnfd_file, vnf_name)

        vnf_id = vnf_instance['vnf']['id']

        def _waiting_time(count):
            self.wait_until_vnf_active(
                vnf_id,
                constants.VNF_CIRROS_CREATE_TIMEOUT,
                constants.ACTIVE_SLEEP_TIME)
            vnf = self.client.show_vnf(vnf_id)['vnf']
            # {"VDU1": ["10.0.0.14", "10.0.0.5"]}
            self.assertEqual(count, len(jsonutils.loads(vnf[
                'mgmt_ip_address'])['VDU1']))

        def _inject_monitoring_policy(vnfd_dict):
            polices = vnfd_dict['topology_template'].get('policies', [])
            mon_policy = dict()
            for policy_dict in polices:
                for name, policy in policy_dict.items():
                    if policy['type'] == constants.POLICY_ALARMING:
                        triggers = policy['triggers']
                        for trigger_name, trigger_dict in triggers.items():
                            policy_action_list = trigger_dict['action']
                            for policy_action_name in policy_action_list:
                                mon_policy[trigger_name] = policy_action_name
            return mon_policy

        def verify_policy(policy_dict, kw_policy):
            for name, action in policy_dict.items():
                if kw_policy in name:
                    return name

        # trigger alarm
        monitoring_policy = _inject_monitoring_policy(tosca_dict)
        for mon_policy_name, mon_policy_action in monitoring_policy.items():
            if mon_policy_action in constants.DEFAULT_ALARM_ACTIONS:
                self.wait_until_vnf_active(
                    vnf_id,
                    constants.VNF_CIRROS_CREATE_TIMEOUT,
                    constants.ACTIVE_SLEEP_TIME)
                self.trigger_vnf(vnf_id, mon_policy_name, mon_policy_action)
            else:
                if 'scaling_out' in mon_policy_name:
                    _waiting_time(2)
                    time.sleep(constants.SCALE_WINDOW_SLEEP_TIME)
                    # scaling-out backend action
                    scaling_out_action = mon_policy_action + '-out'
                    self.trigger_vnf(
                        vnf_id, mon_policy_name, scaling_out_action)

                    _waiting_time(3)

                    scaling_in_name = verify_policy(monitoring_policy,
                                                    kw_policy='scaling_in')
                    if scaling_in_name:
                        time.sleep(constants.SCALE_WINDOW_SLEEP_TIME)
                        # scaling-in backend action
                        scaling_in_action = mon_policy_action + '-in'
                        self.trigger_vnf(
                            vnf_id, scaling_in_name, scaling_in_action)

                        _waiting_time(2)

                    self.verify_vnf_crud_events(
                        vnf_id, evt_constants.RES_EVT_SCALE,
                        evt_constants.ACTIVE, cnt=2)
                    self.verify_vnf_crud_events(
                        vnf_id, evt_constants.RES_EVT_SCALE,
                        evt_constants.PENDING_SCALE_OUT, cnt=1)
                    self.verify_vnf_crud_events(
                        vnf_id, evt_constants.RES_EVT_SCALE,
                        evt_constants.PENDING_SCALE_IN, cnt=1)
        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, ("Failed to delete vnf %s after the monitor test" %
                           vnf_id)

        # Verify VNF monitor events captured for states, ACTIVE and DEAD
        vnf_state_list = [evt_constants.ACTIVE, evt_constants.DEAD]
        self.verify_vnf_monitor_events(vnf_id, vnf_state_list)

        # Wait for delete vnf_instance
        self.addCleanup(self.wait_until_vnf_delete, vnf_id,
                        constants.VNF_CIRROS_DELETE_TIMEOUT)

    @unittest.skip("heat-translator use old metric cpu_util & aggregation API")
    def test_vnf_alarm_respawn(self):
        self._test_vnf_tosca_alarm(
            'sample-tosca-alarm-respawn.yaml',
            'alarm and respawn-vnf')

    @unittest.skip("heat-translator use old metric cpu_util & aggregation API")
    def test_vnf_alarm_scale(self):
        self._test_vnf_tosca_alarm(
            'sample-tosca-alarm-scale.yaml',
            'alarm and scale vnf')



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Context: context for security/db session."""

import copy
import datetime

from castellan.common.credentials import keystone_password
from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_db.sqlalchemy import enginefacade

from tacker.common import exceptions
from tacker.db import api as db_api
from tacker import policy

CONF = cfg.CONF


class ContextBase(oslo_context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """

    def __init__(self, user_id=None, tenant_id=None, is_admin=None,
                 timestamp=None, tenant_name=None, user_name=None,
                 is_advsvc=None, **kwargs):
        # NOTE(jamielennox): We maintain this argument in order for tests that
        # pass arguments positionally.
        kwargs.setdefault('project_id', tenant_id)
        # prefer project_name, as that's what's going to be set by
        # keystone. Fall back to tenant_name if for some reason it's blank.
        kwargs.setdefault('project_name', tenant_name)
        super(ContextBase, self).__init__(
            is_admin=is_admin, user_id=user_id, **kwargs)

        self.user_name = user_name

        if not timestamp:
            timestamp = datetime.datetime.utcnow()
        self.timestamp = timestamp
        # self.is_advsvc = is_advsvc
        # if self.is_advsvc is None:
        #     self.is_advsvc = self.is_admin or policy.check_is_advsvc(self)
        if self.is_admin is None:
            self.is_admin = policy.check_is_admin(self)

    @property
    def tenant_id(self):
        return self.project_id

    @tenant_id.setter
    def tenant_id(self, tenant_id):
        self.project_id = tenant_id

    @property
    def tenant_name(self):
        return self.project_name

    @tenant_name.setter
    def tenant_name(self, tenant_name):
        self.project_name = tenant_name

    def to_dict(self):
        context = super(ContextBase, self).to_dict()
        context.update({
            'user_id': self.user_id,
            'tenant_id': self.tenant_id,
            'project_id': self.project_id,
            'timestamp': str(self.timestamp),
            'tenant_name': self.tenant_name,
            'project_name': self.tenant_name,
            'user_name': self.user_name,
        })
        return context

    @classmethod
    def from_dict(cls, values):
        return cls(user_id=values.get('user_id'),
                   tenant_id=values.get('tenant_id', values.get('project_id')),
                   is_admin=values.get('is_admin'),
                   roles=values.get('roles'),
                   timestamp=values.get('timestamp'),
                   request_id=values.get('request_id'),
                   tenant_name=values.get('tenant_name'),
                   user_name=values.get('user_name'),
                   auth_token=values.get('auth_token'))

    def to_policy_values(self):
        values = super(ContextBase, self).to_policy_values()
        values['tenant_id'] = self.project_id
        values['is_admin'] = self.is_admin

        # NOTE(jamielennox): These are almost certainly unused and non-standard
        # but kept for backwards compatibility. Remove them in Pike
        # (oslo.context from Ocata release already issues deprecation warnings
        # for non-standard keys).
        values['user_id'] = self.user_id
        values['project_id'] = self.project_id
        values['domain'] = self.domain_id
        values['user_domain'] = self.user_domain_id
        values['project_domain'] = self.project_domain_id
        values['tenant_name'] = self.project_name
        values['project_name'] = self.project_name
        values['user_name'] = self.user_name

        return values

    def elevated(self):
        """Return a version of this context with admin flag set."""
        context = copy.copy(self)
        context.is_admin = True

        if 'admin' not in [x.lower() for x in context.roles]:
            context.roles = context.roles + ["admin"]

        return context

    def can(self, action, target=None, fatal=True):
        """Verifies that the given action is valid on the target in this context.

        :param action: string representing the action to be checked.
        :param target: dictionary representing the object of the action
            for object creation this should be a dictionary representing the
            location of the object e.g. ``{'project_id': context.project_id}``.
            If None, then this default target will be considered:
            {'project_id': self.project_id, 'user_id': self.user_id}
        :param fatal: if False, will return False when an exception.Forbidden
           occurs.

        :raises tacker.exception.Forbidden: if verification fails and fatal
            is True.

        :return: returns a non-False value (not necessarily "True") if
            authorized and False if not authorized and fatal is False.
        """
        if target is None:
            target = {'tenant_id': self.tenant_id,
                      'user_id': self.user_id}
        try:
            return policy.authorize(self, action, target)
        except exceptions.Forbidden:
            if fatal:
                raise
            return False


@enginefacade.transaction_context_provider
class ContextBaseWithSession(ContextBase):
    pass


class Context(ContextBaseWithSession):
    def __init__(self, *args, **kwargs):
        super(Context, self).__init__(*args, **kwargs)
        self._session = None
        self._api_version = None

    @property
    def session(self):
        # TODO(akamyshnikova): checking for session attribute won't be needed
        # when reader and writer will be used
        if hasattr(super(Context, self), 'session'):
            return super(Context, self).session
        if self._session is None:
            self._session = db_api.get_session()
        return self._session

    @property
    def api_version(self):
        return self._api_version

    @api_version.setter
    def api_version(self, api_version):
        self._api_version = api_version


def get_admin_context():
    return Context(user_id=None,
                   tenant_id=None,
                   is_admin=True,
                   overwrite=False)


def get_admin_context_without_session():
    return ContextBase(user_id=None,
                       tenant_id=None,
                       is_admin=True)


def is_user_context(context):
    """Indicates if the request context is a normal user."""
    if not context:
        return False
    if context.is_admin:
        return False
    if not context.user_id or not context.project_id:
        return False
    return True


def generate_tacker_service_context():
    return keystone_password.KeystonePassword(
        password=CONF.keystone_authtoken.password,
        auth_url=CONF.keystone_authtoken.auth_url,
        username=CONF.keystone_authtoken.username,
        user_domain_name=CONF.keystone_authtoken.user_domain_name,
        project_name=CONF.keystone_authtoken.project_name,
        project_domain_name=CONF.keystone_authtoken.project_domain_name)



# Copyright (C) 2022 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os
import re
import time
import urllib.request as urllib2
import yaml

from oslo_log import log as logging
from oslo_utils import encodeutils
from tacker.common.container import kubernetes_utils
from tacker.common import exceptions
from tacker.common import log
from tacker import objects
from tacker.vnflcm import utils as vnflcm_utils
from tacker.vnfm.infra_drivers.kubernetes.k8s import translate_outputs
from tacker.vnfm.infra_drivers.kubernetes.kubernetes_driver import Kubernetes
from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver
from toscaparser import tosca_template
from urllib.parse import urlparse

LOG = logging.getLogger(__name__)


class ContainerUpdateMgmtDriver(vnflcm_abstract_driver.
                                VnflcmMgmtAbstractDriver):

    def __init__(self):
        pass

    def get_type(self):
        return "mgmt-container-update"

    def get_name(self):
        return "mgmt-container-update"

    def get_description(self):
        return 'Tacker Container Update VNF Mgmt Driver'

    def modify_information_start(self, context, vnf_instance,
                                 modify_vnf_request=None, **kwargs):
        pass

    def _get_kind_and_name(self, file, vnf_package_path):
        # kind_and_names ----> configmap_secrets list or resources list
        kind_and_names = []
        # Read the contents of the manifest file and get the name and kind
        if ((urlparse(file).scheme == 'file') or
                (bool(urlparse(file).scheme) and
                 bool(urlparse(file).netloc))):
            file_content = urllib2.urlopen(file).read()
        else:
            manifest_file = os.path.join(vnf_package_path, file)
            with open(manifest_file, 'r', encoding='utf-8') as f:
                file_content = f.read()
        file_contents = yaml.safe_load_all(file_content)
        for file_content in file_contents:
            kind = file_content.get('kind', '')
            name = file_content.get('metadata', {}).get('name', '')
            kind_and_names.append({'kind': kind, 'name': name})
        return kind_and_names

    def _initialize_k8s_client(self, auth_cred):
        k8s_clients = (kubernetes_utils.KubernetesHTTPAPI().
                       get_k8s_client_dict(auth_cred))
        return k8s_clients

    def _get_k8s_objs(self, target_k8s_files, vnf_package_path, namespace,
                      k8s_clients):
        transformer = translate_outputs.Transformer(
            None, None, None, k8s_clients)
        k8s_objs = transformer.get_k8s_objs_from_yaml(
            target_k8s_files, vnf_package_path, namespace)
        return k8s_objs

    def _modify_container_img(self, old_containers, new_containers):
        for old_container in old_containers:
            for new_container in new_containers:
                if (old_container.name == new_container.name and
                        old_container.image != new_container.image):
                    # Replace the old image with the new image
                    old_container.image = new_container.image
                    break

    def _replace_api(self, k8s_clients, namespace, k8s_obj):
        # get api
        name = k8s_obj['object'].metadata.name
        kind = k8s_obj['object'].kind
        api_version = k8s_obj['object'].api_version
        body = k8s_obj['object']

        def convert(name):
            name_with_underscores = re.sub(
                '(.)([A-Z][a-z]+)', r'\1_\2', name)
            return re.sub('([a-z0-9])([A-Z])', r'\1_\2',
                          name_with_underscores).lower()

        snake_case_kind = convert(kind)
        try:
            replace_api = eval(f'k8s_clients["{api_version}"].'
                               f'replace_namespaced_{snake_case_kind}')
            response = replace_api(name=name, namespace=namespace,
                                   body=body)
        except Exception as exp:
            raise exceptions.MgmtDriverOtherError(
                error_message=encodeutils.exception_to_unicode(exp))

        return response

    def _replace_wait_k8s(self, kube_driver, k8s_pod_objs,
                          core_v1_api_client, vnf_instance):
        try:
            time.sleep(kube_driver.STACK_RETRY_WAIT)
            status = 'Pending'
            stack_retries = kube_driver.STACK_RETRIES

            while status == 'Pending' and stack_retries > 0:
                pods_information = []
                for k8s_pod_obj in k8s_pod_objs:
                    kind = k8s_pod_obj['object'].kind
                    namespace = k8s_pod_obj.get('namespace')
                    if k8s_pod_obj['object'].metadata:
                        name = k8s_pod_obj['object'].metadata.name
                    else:
                        name = ''

                    response = core_v1_api_client.list_namespaced_pod(
                        namespace=namespace)
                    for pod in response.items:
                        match_result = kube_driver.is_match_pod_naming_rule(
                            kind, name, pod.metadata.name)
                        if match_result:
                            pods_information.append(pod)
                status = kube_driver.get_pod_status(pods_information)
                if status == 'Unknown':
                    wait = (kube_driver.STACK_RETRIES *
                            kube_driver.STACK_RETRY_WAIT)
                    error_reason = (
                        f"Resource creation is not completed within"
                        f" {wait} seconds as creation of CNF "
                        f"{vnf_instance.id} is not completed")
                    raise exceptions.MgmtDriverOtherError(
                        error_message=error_reason)
                if status == 'Pending':
                    stack_retries = stack_retries - 1
                    time.sleep(kube_driver.STACK_RETRY_WAIT)
            if stack_retries == 0 and status != 'Running':
                LOG.error('It is time out, When modify cnf,'
                          'waiting for resource creation.')
                wait = (kube_driver.STACK_RETRIES *
                        kube_driver.STACK_RETRY_WAIT)
                error_reason = (f"Resource creation is not completed within"
                                f" {wait} seconds as creation of CNF "
                                f"{vnf_instance.id} is not completed")
                raise exceptions.MgmtDriverOtherError(
                    error_message=error_reason)
            return k8s_pod_objs
        except Exception as exp:
            raise exceptions.MgmtDriverOtherError(
                error_message=encodeutils.exception_to_unicode(exp))

    @log.log
    def modify_information_end(self, context, vnf_instance,
                               modify_vnf_request=None, **kwargs):
        kube_driver = Kubernetes()
        # get old_vnf_package_path
        old_vnf_package_path = kwargs['old_vnf_package_path']

        # get configmap_secret_paths
        configmap_secret_paths = kwargs['configmap_secret_paths']
        # Get the path of the VNF Package according to vnf_instance.vnfd_id
        new_vnf_package_path = vnflcm_utils.get_vnf_package_path(
            context, vnf_instance.vnfd_id)

        # Get the input parameters of instantiate
        inst_vnf_info = vnf_instance.instantiated_vnf_info

        # Make a deep copy of inst_vnf_info
        new_inst_vnf_info = copy.deepcopy(inst_vnf_info)

        # Get vim_connection_info from vnf_instance
        vim_info = vnflcm_utils.get_vim(context,
                                        vnf_instance.vim_connection_info)
        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)
        for configmap_secret_path in configmap_secret_paths:
            # Read the contents of the manifest file and get name and kind
            configmap_secrets = self._get_kind_and_name(configmap_secret_path,
                                                        new_vnf_package_path)
            for index, manifest_path in enumerate(
                    inst_vnf_info.additional_params[
                        'lcm-kubernetes-def-files']):
                # Read the contents of the manifest file and get name and kind
                resources = self._get_kind_and_name(manifest_path,
                                                    old_vnf_package_path)
                resource = [obj for obj in resources if
                            obj in configmap_secrets]
                if resource:
                    if len(resource) == len(configmap_secrets):
                        # Update the manifest file path of ConfigMap/Secret
                        new_inst_vnf_info.additional_params[
                            'lcm-kubernetes-def-files'][
                            index] = configmap_secret_path
                        break
                    raise exceptions.MgmtDriverOtherError(
                        error_message='The number of resources in the '
                                      'manifest file of the changed '
                                      'ConfigMap/Secret is inconsistent '
                                      'with the previous one.')
        # Initialize k8s client api
        auth_attr = vim_connection_info.access_info
        auth_cred, _ = kube_driver.get_auth_creds(auth_attr)
        k8s_clients = self._initialize_k8s_client(auth_cred)

        # Get the namespace of this CNF
        namespace = vnf_instance.vnf_metadata['namespace']

        # Get old_k8s_objs
        target_k8s_files = inst_vnf_info.additional_params[
            'lcm-kubernetes-def-files']
        old_k8s_objs = self._get_k8s_objs(target_k8s_files,
                                          old_vnf_package_path, namespace,
                                          k8s_clients)

        # Get new_k8s_objs
        target_k8s_files = new_inst_vnf_info.additional_params[
            'lcm-kubernetes-def-files']
        new_k8s_objs = self._get_k8s_objs(target_k8s_files,
                                          new_vnf_package_path, namespace,
                                          k8s_clients)
        # Initialize k8s_pod_objs and k8s_config_objs
        k8s_pod_objs = []
        k8s_config_objs = []

        for old_k8s_obj in old_k8s_objs:
            old_k8s_obj_kind = old_k8s_obj['object'].kind
            old_k8s_obj_name = old_k8s_obj['object'].metadata.name
            if old_k8s_obj_kind in ['Pod', 'Deployment']:
                for new_k8s_obj in new_k8s_objs:
                    # If the old and new k8s_obj have the same kind and name
                    new_k8s_obj_kind = new_k8s_obj['object'].kind
                    new_k8s_obj_name = new_k8s_obj['object'].metadata.name
                    if old_k8s_obj_kind == new_k8s_obj_kind and (
                            old_k8s_obj_name == new_k8s_obj_name):
                        # Call the read API
                        old_k8s_resource_info = (
                            kube_driver.select_k8s_obj_read_api(
                                k8s_client_dict=k8s_clients,
                                namespace=namespace,
                                name=old_k8s_obj_name,
                                kind=old_k8s_obj_kind,
                                api_version=old_k8s_obj['object'].api_version
                            ))
                        # Assign old_k8s_resource_info to old_k8s_obj['object']
                        old_k8s_obj['object'] = old_k8s_resource_info

                        if old_k8s_obj_kind == 'Deployment':
                            old_containers = (old_k8s_obj['object'].spec
                                              .template.spec.containers)
                            new_containers = (new_k8s_obj['object'].spec
                                              .template.spec.containers)
                        elif old_k8s_obj_kind == 'Pod':
                            old_containers = (old_k8s_obj['object']
                                              .spec.containers)
                            new_containers = (new_k8s_obj['object']
                                              .spec.containers)
                        # Replace the old image with the new image
                        self._modify_container_img(old_containers,
                                                   new_containers)
                        break

                # Append old_k8s_obj to k8s_pod_objs
                k8s_pod_objs.append(old_k8s_obj)
            elif old_k8s_obj_kind in ['ConfigMap', 'Secret']:
                for new_k8s_obj in new_k8s_objs:
                    # If the old and new k8s_obj have the same kind and name
                    new_k8s_obj_kind = new_k8s_obj['object'].kind
                    new_k8s_obj_name = new_k8s_obj['object'].metadata.name
                    if old_k8s_obj_kind == new_k8s_obj_kind and (
                            old_k8s_obj_name == new_k8s_obj_name):
                        # Append old_k8s_obj to k8s_pod_objs
                        k8s_config_objs.append(new_k8s_obj)
                        break
        for k8s_config_obj in k8s_config_objs:
            # Call the replace API
            self._replace_api(k8s_clients, namespace, k8s_config_obj)
        for k8s_pod_obj in k8s_pod_objs:
            # Call the replace API
            self._replace_api(k8s_clients, namespace, k8s_pod_obj)

        # _replace_wait_k8s
        core_v1_api_client = k8s_clients['v1']
        self._replace_wait_k8s(kube_driver, k8s_pod_objs, core_v1_api_client,
                               vnf_instance)
        # Get all pod information under the specified namespace
        pods = core_v1_api_client.list_namespaced_pod(namespace=namespace)

        # get TOSCA node templates
        vnfd_dict = vnflcm_utils.get_vnfd_dict(
            context, vnf_instance.vnfd_id,
            vnf_instance.instantiated_vnf_info.flavour_id)
        tosca = tosca_template.ToscaTemplate(
            parsed_params={}, a_file=False, yaml_dict_tpl=vnfd_dict)
        tosca_node_tpls = tosca.topology_template.nodetemplates
        # get vdu_names dict {vdu_id: vdu_name}
        vdu_names = {}
        for node_tpl in tosca_node_tpls:
            for node_name, node_value in node_tpl.templates.items():
                if node_value.get('type') == "tosca.nodes.nfv.Vdu.Compute":
                    vdu_id = node_name
                    vdu_name = node_value.get('properties').get('name')
                    vdu_names[vdu_id] = vdu_name

        for vnfc_resource in new_inst_vnf_info.vnfc_resource_info:
            for pod in pods.items:
                # the name of the pod matches the resource_id in vnfc_resource
                match_result = kube_driver.is_match_pod_naming_rule(
                    vnfc_resource.compute_resource.vim_level_resource_type,
                    vdu_names[vnfc_resource.vdu_id],
                    pod.metadata.name)
                if match_result:
                    # Update the name of the pod in the resourceId
                    vnfc_resource.compute_resource.resource_id = (
                        pod.metadata.name)
                    # Delete the current pod from pods.items
                    pods.items.remove(pod)
                    break
        # the ConfigMap/Secret file path in inst_vnf_info.additional_params
        vnf_instance.instantiated_vnf_info.vnfc_resource_info = (
            new_inst_vnf_info.vnfc_resource_info)
        vnf_instance.instantiated_vnf_info.additional_params = (
            new_inst_vnf_info.additional_params)
        vnf_instance.save()

    def instantiate_start(self, context, vnf_instance,
                          instantiate_vnf_request, grant,
                          grant_request, **kwargs):
        pass

    def instantiate_end(self, context, vnf_instance,
                        instantiate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    def terminate_start(self, context, vnf_instance,
                        terminate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    def terminate_end(self, context, vnf_instance,
                      terminate_vnf_request, grant,
                      grant_request, **kwargs):
        pass

    def scale_start(self, context, vnf_instance,
                    scale_vnf_request, grant,
                    grant_request, **kwargs):
        pass

    def scale_end(self, context, vnf_instance,
                  scale_vnf_request, grant,
                  grant_request, **kwargs):
        pass

    def heal_start(self, context, vnf_instance,
                   heal_vnf_request, grant,
                   grant_request, **kwargs):
        pass

    def heal_end(self, context, vnf_instance,
                 heal_vnf_request, grant,
                 grant_request, **kwargs):
        pass

    def change_external_connectivity_start(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    def change_external_connectivity_end(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions


class NFVPluginBase(extensions.PluginInterface, metaclass=abc.ABCMeta):
    """Define base interface for any Advanced Service plugin."""
    supported_extension_aliases = []

    @abc.abstractmethod
    def get_plugin_type(self):
        """Return one of predefined service types.

        See tacker/plugins/common/constants.py
        """
        pass

    @abc.abstractmethod
    def get_plugin_name(self):
        """Return a symbolic name for the plugin.

        Each service plugin should have a symbolic name. This name
        will be used, for instance, by service definitions in service types
        """
        pass

    @abc.abstractmethod
    def get_plugin_description(self):
        """Return string description of the plugin."""
        pass



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import yaml

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from tacker.tests import uuidsentinel


class Grant:
    GRANT_REQ_PATH = '/grant/v1/grants'

    ZONES = [
        {
            "id": uuidsentinel.zone_id,
            "zoneId": "nova",
            "vimConnectionId": uuidsentinel.vim_connection_id
        }
    ]

    ADDITIONAL_PARAMS = {
        "key": "value"
    }

    @staticmethod
    def _make_vim_connection_info(tenant_id):
        access_info = {
            "username": "nfv_user",
            "region": "RegionOne",
            "password": "devstack",
            "tenant": tenant_id
        }

        return [{
            "id": uuidsentinel.vim_connection_id,
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.v_2",
            "interfaceInfo": {
                "endpoint": "http://127.0.0.1/identity"
            },
            "accessInfo": access_info
        }]

    @staticmethod
    def _make_add_resources(req_add_resources, zones=None,
                            placement_constraints=None):
        add_resources = []
        for req_add_resource in req_add_resources:
            res_add_resource = {
                "resourceDefinitionId": req_add_resource['id'],
                "vimConnectionId": uuidsentinel.vim_connection_id
            }
            if req_add_resource['type'] == 'COMPUTE':
                if placement_constraints:
                    zone_id_dict = (Grant.
                                    _get_zone_id_from_placement_constraint(
                                        placement_constraints, zones))
                    if zone_id_dict[req_add_resource['id']]:
                        res_add_resource["zoneId"] = zone_id_dict[
                            req_add_resource['id']]
            add_resources.append(res_add_resource)

        return add_resources

    @staticmethod
    def _get_zone_id_from_placement_constraint(placement_constraints, zones):
        zone_id_dict = {}
        for placement_constraint in placement_constraints:
            for index in range(len(placement_constraint['resource'])):
                if placement_constraint[
                        'affinityOrAntiAffinity'] == 'AFFINITY':
                    # In this fake_grant, if the user set 'AFFINITY'
                    # rule in VNF Package, it will always use the first
                    # `Availability Zone` to create VM.
                    zone_id_dict[placement_constraint[
                        'resource'][index]['resourceId']] = zones[0]['id']
                else:
                    try:
                        zone_id_dict[placement_constraint[
                            'resource'][index]['resourceId']] = zones[
                            index]['id']
                    except IndexError:
                        print(
                            "The number of 'Availability Zone'"
                            "cannot support current case.")
                        raise IndexError

        return zone_id_dict

    @staticmethod
    def _make_zones(zone_name_list):
        zone = []
        for name in zone_name_list:
            zone_dict = {
                "id": uuidutils.generate_uuid(),
                "zoneId": name,
                "vimConnectionId": uuidsentinel.vim_connection_id
            }
            zone.append(zone_dict)
        return zone

    @staticmethod
    def _make_remove_resources(req_remove_resources):
        res_remove_resources = []
        for req_remove_resource in req_remove_resources:
            res_remove_resource = {
                "resourceDefinitionId": req_remove_resource['id']
            }

            res_remove_resources.append(res_remove_resource)

        return res_remove_resources

    @staticmethod
    def _make_update_resources(req_update_resources,
                               zones=None, placement_constraints=None):
        res_update_resources = []
        for req_update_resource in req_update_resources:
            res_update_resource = {
                "resourceDefinitionId": req_update_resource['id'],
                "vimConnectionId": uuidsentinel.vim_connection_id
            }
            if req_update_resource['type'] == 'COMPUTE':
                if placement_constraints:
                    zone_id_dict = (Grant.
                                    _get_zone_id_from_placement_constraint(
                                        placement_constraints, zones))
                    if zone_id_dict[req_update_resource['id']]:
                        res_update_resource["zoneId"] = zone_id_dict[
                            req_update_resource['id']]
            res_update_resources.append(res_update_resource)

        return res_update_resources

    @staticmethod
    def get_vdu_list(add_resources):
        vdu_list = []
        for add_resource in add_resources:
            if add_resource['type'] == 'COMPUTE':
                vdu_list.append(add_resource['vduId'])
        return vdu_list

    @staticmethod
    def _make_vim_assets(add_resources, image_id_dict, flavour_id_dict):
        # set m1.tiny="1" for flavour_id
        vdu_list = Grant.get_vdu_list(add_resources)
        flavors = [Grant._generate_flavour(vdu, flavour_id_dict)
                   for vdu in vdu_list if flavour_id_dict.get(vdu)]
        images = [Grant._generate_image(vdu, image_id_dict)
                  for vdu in vdu_list if image_id_dict.get(vdu)]

        vim_assets = {
            'computeResourceFlavours': flavors,
            'softwareImages': images,
        }

        return vim_assets

    @staticmethod
    def _generate_flavour(vdu, flavour_id_dict):
        if flavour_id_dict.get(vdu):
            return {
                "vimConnectionId": uuidsentinel.vim_connection_id,
                "vnfdVirtualComputeDescId": vdu,
                "vimFlavourId": flavour_id_dict[vdu]
            }
        return None

    @staticmethod
    def _generate_image(vdu, image_id_dict):
        if image_id_dict.get(vdu):
            return {
                "vimConnectionId": uuidsentinel.vim_connection_id,
                "vnfdSoftwareImageId": vdu,
                "vimSoftwareImageId": image_id_dict[vdu]
            }
        return None

    @staticmethod
    def _make_response_template(request_body):
        res = {
            "id": uuidsentinel.__getattr__(request_body['vnfLcmOpOccId']),
            "vnfInstanceId": request_body['vnfInstanceId'],
            "vnfLcmOpOccId": request_body['vnfLcmOpOccId'],
        }
        res["_links"] = {
            "self": {
                # set fake server port.
                "href": os.path.join(
                    'http://localhost:9990',
                    Grant.GRANT_REQ_PATH)},
            "vnfLcmOpOcc": {
                "href": os.path.join(
                    'http://localhost:9890/vnflcm/v1/vnf_lcm_op_occs',
                    request_body['vnfLcmOpOccId'])},
            "vnfInstance": {
                "href": os.path.join(
                    'http://localhost:9890/vnflcm/v1/vnf_instances',
                    request_body['vnfInstanceId'])}}

        return res

    @staticmethod
    def convert_body_to_dict(body):
        if isinstance(body, str):
            return jsonutils.loads(body)

        return body

    @staticmethod
    def make_inst_response_body(
            request_body, tenant_id, image_id_dict,
            flavour_id_dict, zone_name_list):
        request_body = Grant.convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant._make_zones(zone_name_list)
        if 'addResources' in request_body.keys():
            res["addResources"] = Grant._make_add_resources(
                request_body['addResources'], res["zones"],
                request_body.get('placementConstraints'))
            res["vimAssets"] = Grant._make_vim_assets(
                request_body['addResources'],
                image_id_dict,
                flavour_id_dict)
        res["additionalParams"] = Grant.ADDITIONAL_PARAMS

        return res

    @staticmethod
    def make_heal_response_body(request_body, tenant_id, image_id_dict,
                                flavour_id_dict, zone_name_list):
        request_body = Grant.convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant._make_zones(zone_name_list)
        if 'addResources' in request_body.keys():
            res["addResources"] = Grant._make_add_resources(
                request_body['addResources'], res["zones"],
                request_body.get('placementConstraints'))
            res["vimAssets"] = Grant._make_vim_assets(
                request_body['addResources'],
                image_id_dict,
                flavour_id_dict)
        if 'removeResources' in request_body.keys():
            res["removeResources"] = Grant._make_remove_resources(
                request_body['removeResources'])
        if 'updateResources' in request_body.keys():
            res["updateResources"] = Grant._make_update_resources(
                request_body['updateResources'], res["zones"],
                request_body.get('placementConstraints'))
        return res

    @staticmethod
    def make_scale_response_body(
            request_body, tenant_id, image_id_dict, flavour_id_dict,
            zone_name_list):
        request_body = Grant.convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        if 'addResources' in request_body.keys():
            res["zones"] = Grant._make_zones(zone_name_list)
            res["addResources"] = Grant._make_add_resources(
                request_body['addResources'], res["zones"],
                request_body.get('placementConstraints'))
            res["vimAssets"] = Grant._make_vim_assets(
                request_body['addResources'],
                image_id_dict,
                flavour_id_dict)
            res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        if 'removeResources' in request_body.keys():
            res["removeResources"] = Grant._make_remove_resources(
                request_body['removeResources'])

        return res

    @staticmethod
    def make_term_response_body(request_body):
        request_body = Grant.convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        if 'removeResources' in request_body.keys():
            res["removeResources"] = Grant._make_remove_resources(
                request_body['removeResources'])

        return res

    @staticmethod
    def make_change_ext_conn_response_body(
            request_body, tenant_id, zone_name_list):
        request_body = Grant.convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant._make_zones(zone_name_list)
        if 'updateResources' in request_body.keys():
            res["updateResources"] = Grant._make_update_resources(
                request_body['updateResources'], res["zones"],
                request_body.get('placementConstraints'))
        res["additionalParams"] = Grant.ADDITIONAL_PARAMS

        return res

    @staticmethod
    def get_sw_image(package_dir="functional6"):
        sample_name = package_dir
        csar_package_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../../etc/samples/etsi/nfv",
                sample_name))
        yaml_file = os.path.join(csar_package_path,
                                 "Definitions/helloworld3_df_simple.yaml")
        with open(yaml_file, 'r', encoding='utf-8') as f:
            config_content = yaml.safe_load(f.read())

        nodes = (config_content
                 .get('topology_template', {})
                 .get('node_templates', {}))
        types = ['tosca.nodes.nfv.Vdu.Compute',
                 'tosca.nodes.nfv.Vdu.VirtualBlockStorage']
        sw_image = {}
        for name, data in nodes.items():
            if (data['type'] in types and
                    data.get('properties', {}).get('sw_image_data')):
                image = data['properties']['sw_image_data']['name']
                sw_image[name] = image

        return sw_image

    @staticmethod
    def get_compute_flavor(package_dir="functional6"):
        sample_name = package_dir
        csar_package_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../../etc/samples/etsi/nfv",
                sample_name))
        yaml_file = os.path.join(csar_package_path,
                                 "Definitions/helloworld3_df_simple.yaml")
        with open(yaml_file, 'r', encoding='utf-8') as f:
            config_content = yaml.safe_load(f.read())

        nodes = (config_content
                 .get('topology_template', {})
                 .get('node_templates', {}))
        types = ['tosca.nodes.nfv.Vdu.Compute',
                 'tosca.nodes.nfv.Vdu.VirtualBlockStorage']
        flavor = {}
        for name, data in nodes.items():
            if (data['type'] in types and
                    data.get('capabilities', {})
                        .get('virtual_compute', {})
                        .get('properties', {})
                        .get('requested_additional_capabilities', {})
                        .get('properties')):
                flavour = data['capabilities']['virtual_compute'][
                    'properties']['requested_additional_capabilities'][
                    'properties']['requested_additional_capability_name']
                flavor[name] = flavour

        return flavor



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re
import toscaparser.utils.yamlparser
from urllib.parse import urlparse
import urllib.request as urllib2
import yaml

from kubernetes import client
from oslo_config import cfg
from oslo_log import log as logging
from tacker.common import exceptions


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

YAML_LOADER = toscaparser.utils.yamlparser.load_yaml
NEWLINE_CHARACTER = "\n"
COLON_CHARACTER = ':'
WHITE_SPACE_CHARACTER = ' '
NON_WHITE_SPACE_CHARACTER = ''
HYPHEN_CHARACTER = '-'
DASH_CHARACTER = '_'
# Due to the dependency of k8s resource creation, according to the design,
# other resources (resources not mentioned in self.RESOURCE_CREATION_SORT)
# will be created after the NetworkPolicy resource. This number is a flag
# to ensure that when multiple resources are to be created, the order of
# other resources is after NetworkPolicy and before Service.
OTHER_RESOURCE_SORT_POSITION = 8


class Transformer(object):
    """Transform TOSCA template to Kubernetes resources"""

    def __init__(self, core_v1_api_client, app_v1_api_client,
                 scaling_api_client, k8s_client_dict):
        # the old param used when creating vnf with TOSCA template
        self.core_v1_api_client = core_v1_api_client
        self.app_v1_api_client = app_v1_api_client
        self.scaling_api_client = scaling_api_client
        # the new param used when instantiating vnf with addtionalParams
        self.k8s_client_dict = k8s_client_dict
        self.RESOURCE_CREATION_ORDER = [
            'StorageClass',
            'PersistentVolume',
            'PriorityClass',
            'Namespace',
            'LimitRange',
            'ResourceQuota',
            'HorizontalPodAutoscaler',
            'NetworkPolicy',
            'Service',
            'Endpoints',
            'PersistentVolumeClaim',
            'ConfigMap',
            'Secret',
            'StatefulSet',
            'Job',
            'Deployment',
            'DaemonSet',
            'Pod'
        ]
        self.method_value = {
            "Pod": 'create_namespaced_pod',
            "Service": 'create_namespaced_service',
            "ConfigMap": 'create_namespaced_config_map',
            "Secret": 'create_namespaced_secret',
            "PersistentVolumeClaim":
                'create_namespaced_persistent_volume_claim',
            "LimitRange": 'create_namespaced_limit_range',
            "PodTemplate": 'create_namespaced_pod_template',
            "Binding": 'create_namespaced_binding',
            "Namespace": 'create_namespace',
            "Node": 'create_node',
            "PersistentVolume": 'create_persistent_volume',
            "ResourceQuota": 'create_namespaced_resource_quota',
            "ServiceAccount": 'create_namespaced_service_account',
            "APIService": 'create_api_service',
            "DaemonSet": 'create_namespaced_daemon_set',
            "Deployment": 'create_namespaced_deployment',
            "ReplicaSet": 'create_namespaced_replica_set',
            "StatefulSet": 'create_namespaced_stateful_set',
            "ControllerRevision": 'create_namespaced_controller_revision',
            "TokenReview": 'create_token_review',
            "LocalSubjectAccessReview": 'create_namespaced_local_'
                                        'subject_access_review',
            "SelfSubjectAccessReview": 'create_self_subject_access_review',
            "SelfSubjectRulesReview": 'create_self_subject_rules_review',
            "SubjectAccessReview": 'create_subject_access_review',
            "HorizontalPodAutoscaler": 'create_namespaced_horizontal_'
                                       'pod_autoscaler',
            "Job": 'create_namespaced_job',
            "Lease": 'create_namespaced_lease',
            "NetworkPolicy": 'create_namespaced_network_policy',
            "ClusterRole": 'create_cluster_role',
            "ClusterRoleBinding": 'create_cluster_role_binding',
            "Role": 'create_namespaced_role',
            "RoleBinding": 'create_namespaced_role_binding',
            "PriorityClass": 'create_priority_class',
            "StorageClass": 'create_storage_class',
            "VolumeAttachment": 'create_volume_attachment',
        }

    def transform(self, tosca_kube_objects):
        """transform function translates from tosca_kube_object to

        kubernetes_object (ConfigMap, Deployment, Service, HPA)
        """

        # kubernetes_objects store all kubernetes objects that are transformed
        # from TOSCA VNF template
        kubernetes_objects = dict()
        for tosca_kube_obj in tosca_kube_objects:
            namespace = tosca_kube_obj.namespace
            kubernetes_objects['namespace'] = namespace
            kubernetes_objects['objects'] = list()
            kube_obj_name = tosca_kube_obj.name
            new_kube_obj_name = self.pre_process_name(kube_obj_name)

            # translate environments to ConfigMap objects
            for container in tosca_kube_obj.containers:
                config_map_object = \
                    self.init_configmap(container_props=container,
                                        kube_obj_name=new_kube_obj_name)
                if config_map_object:
                    kubernetes_objects['objects'].append(config_map_object)

            # translate Deployment object
            deployment_object = \
                self.init_deployment(tosca_kube_obj=tosca_kube_obj,
                                     kube_obj_name=new_kube_obj_name)
            kubernetes_objects['objects'].append(deployment_object)

            # translate to Horizontal Pod Autoscaler object
            hpa_object = self.init_hpa(tosca_kube_obj=tosca_kube_obj,
                                       kube_obj_name=new_kube_obj_name)
            if hpa_object:
                kubernetes_objects['objects'].append(hpa_object)

            # translate to Service object
            service_object = self.init_service(
                tosca_kube_obj=tosca_kube_obj,
                kube_obj_name=new_kube_obj_name)
            kubernetes_objects['objects'].append(service_object)

        return kubernetes_objects

    def _create_k8s_object(self, kind, file_content_dict):
        # must_param referring K8s official object page
        # *e.g:https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Service.md
        # initiating k8s object, you need to
        # give the must param an empty value.
        must_param = {
            'V1LocalSubjectAccessReview': '(spec="")',
            'V1HTTPGetAction': '(port="")',
            'V1DeploymentSpec': '(selector="", template="")',
            'V1PodSpec': '(containers=[])',
            'V1ConfigMapKeySelector': '(key="")',
            'V1Container': '(name="")',
            'V1EnvVar': '(name="")',
            'V1SecretKeySelector': '(key="")',
            'V1ContainerPort': '(container_port="")',
            'V1VolumeMount': '(mount_path="", name="")',
            'V1PodCondition': '(status="", type="")',
            'V1ContainerStatus': '('
                                 'image="", image_id="", '
                                 'name="", ready="", '
                                 'restart_count="")',
            'V1ServicePort': '(port=80)',
            'V1TypedLocalObjectReference': '(kind="", name="")',
            'V1LabelSelectorRequirement': '(key="", operator="")',
            'V1PersistentVolumeClaimCondition': '(status="", type="")',
            'V1AWSElasticBlockStoreVolumeSource': '(volume_id="")',
            'V1AzureDiskVolumeSource': '(disk_name="", disk_uri="")',
            'V1AzureFileVolumeSource': '(secret_name="", share_name="")',
            'V1CephFSVolumeSource': '(monitors=[])',
            'V1CinderVolumeSource': '(volume_id="")',
            'V1KeyToPath': '(key="", path="")',
            'V1CSIVolumeSource': '(driver="")',
            'V1DownwardAPIVolumeFile': '(path="")',
            'V1ObjectFieldSelector': '(field_path="")',
            'V1ResourceFieldSelector': '(resource="")',
            'V1FlexVolumeSource': '(driver="")',
            'V1GCEPersistentDiskVolumeSource': '(pd_name="")',
            'V1GitRepoVolumeSource': '(repository="")',
            'V1GlusterfsVolumeSource': '(endpoints="", path="")',
            'V1HostPathVolumeSource': '(path="")',
            'V1ISCSIVolumeSource': '(iqn="", lun=0, target_portal="")',
            'V1Volume': '(name="")',
            'V1NFSVolumeSource': '(path="", server="")',
            'V1PersistentVolumeClaimVolumeSource': '(claim_name="")',
            'V1PhotonPersistentDiskVolumeSource': '(pd_id="")',
            'V1PortworxVolumeSource': '(volume_id="")',
            'V1ProjectedVolumeSource': '(sources=[])',
            'V1ServiceAccountTokenProjection': '(path="")',
            'V1QuobyteVolumeSource': '(registry="", volume="")',
            'V1RBDVolumeSource': '(image="", monitors=[])',
            'V1ScaleIOVolumeSource': '('
                                     'gateway="", secret_ref="", '
                                     'system="")',
            'V1VsphereVirtualDiskVolumeSource': '(volume_path="")',
            'V1LimitRangeSpec': '(limits=[])',
            'V1Binding': '(target="")',
            'V1ComponentCondition': '(status="", type="")',
            'V1NamespaceCondition': '(status="", type="")',
            'V1ConfigMapNodeConfigSource': '(kubelet_config_key="", '
                                           'name="", namespace="")',
            'V1Taint': '(effect="", key="")',
            'V1NodeAddress': '(address="", type="")',
            'V1NodeCondition': '(status="", type="")',
            'V1DaemonEndpoint': '(port=0)',
            'V1ContainerImage': '(names=[])',
            'V1NodeSystemInfo': '(architecture="", boot_id="", '
                                'container_runtime_version="",'
                                'kernel_version="", '
                                'kube_proxy_version="", '
                                'kubelet_version="",'
                                'machine_id="", operating_system="", '
                                'os_image="", system_uuid="")',
            'V1AttachedVolume': '(device_path="", name="")',
            'V1ScopedResourceSelectorRequirement':
                '(operator="", scope_name="")',
            'V1APIServiceSpec': '(group_priority_minimum=0, '
                                'service="", '
                                'version_priority=0)',
            'V1APIServiceCondition': '(status="", type="")',
            'V1DaemonSetSpec': '(selector="", template="")',
            'V1ReplicaSetSpec': '(selector="")',
            'V1StatefulSetSpec': '(selector="", '
                                 'service_name="", template="")',
            'V1StatefulSetCondition': '(status="", type="")',
            'V1StatefulSetStatus': '(replicas=0)',
            'V1ControllerRevision': '(revision=0)',
            'V1TokenReview': '(spec="")',
            'V1SubjectAccessReviewStatus': '(allowed=True)',
            'V1SelfSubjectAccessReview': '(spec="")',
            'V1SelfSubjectRulesReview': '(spec="")',
            'V1SubjectRulesReviewStatus': '(incomplete=True, '
                                          'non_resource_rules=[], '
                                          'resource_rules=[])',
            'V1NonResourceRule': '(verbs=[])',
            'V1SubjectAccessReview': '(spec="")',
            'V1HorizontalPodAutoscalerSpec':
                '(max_replicas=0, scale_target_ref="")',
            'V1CrossVersionObjectReference': '(kind="", name="")',
            'V1HorizontalPodAutoscalerStatus':
                '(current_replicas=0, desired_replicas=0)',
            'V1JobSpec': '(template="")',
            'V1NetworkPolicySpec': '(pod_selector="")',
            'V1PolicyRule': '(verbs=[])',
            'V1ClusterRoleBinding': '(role_ref="")',
            'V1RoleRef': '(api_group="", kind="", name="")',
            'V1Subject': '(kind="", name="")',
            'V1RoleBinding': '(role_ref="")',
            'V1PriorityClass': '(value=0)',
            'V1StorageClass': '(provisioner="")',
            'V1TopologySelectorLabelRequirement': '(key="", values=[])',
            'V1VolumeAttachment': '(spec="")',
            'V1VolumeAttachmentSpec':
                '(attacher="", node_name="", source="")',
            'V1VolumeAttachmentStatus': '(attached=True)',
            'V1NodeSelector': '(node_selector_terms=[])',
            'V1NodeSelectorRequirement': '(key="", operator="")',
            'V1PreferredSchedulingTerm': '(preference="", weight=1)',
            'V1PodAffinityTerm': '(topology_key="")',
            'V1WeightedPodAffinityTerm': '(pod_affinity_term="", weight=1)',
            'V1OwnerReference': '(api_version="", kind="", name="", uid="")',
            'V1HTTPHeader': '(name="", value="")',
            'V1TCPSocketAction': '(port="")',
            'V1VolumeDevice': '(device_path="", name="")',
            'V1PodReadinessGate': '(condition_type="")',
            'V1Sysctl': '(name="", value="")',
            'V1ContainerStateTerminated': '(exit_code=0)',
            'V1AzureFilePersistentVolumeSource': '(secret_name="",'
                                                 ' share_name="")',
            'V1CephFSPersistentVolumeSource': '(monitors=[])',
            'V1CinderPersistentVolumeSource': '(volume_id="")',
            'V1CSIPersistentVolumeSource': '(driver="", volume_handle="")',
            'V1FlexPersistentVolumeSource': '(driver="")',
            'V1GlusterfsPersistentVolumeSource': '(endpoints="", path="")',
            'V1ISCSIPersistentVolumeSource': '(iqn="", lun=0,'
                                             ' target_portal="")',
            'V1LocalVolumeSource': '(path="")',
            'V1RBDPersistentVolumeSource': '(image="", monitors=[])',
            'V1ScaleIOPersistentVolumeSource': '('
                                               'gateway="",'
                                               ' secret_ref="",'
                                               ' system="")',
            'V1DaemonSetStatus': '(current_number_scheduled=0, '
                                 'desired_number_scheduled=0, '
                                 'number_misscheduled=0, '
                                 'number_ready=0)',
            'V1DaemonSetCondition': '(status="", type="")',
            'V1DeploymentCondition': '(status="", type="")',
            'V1ReplicaSetStatus': '(replicas=0)',
            'V1ReplicaSetCondition': '(status="", type="")',
            'V1ResourceRule': '(verbs=[])',
            'V1JobCondition': '(status="", type="")',
            'V1IPBlock': '(cidr="")',
            'V1EphemeralContainer': '(name="")',
            'V1TopologySpreadConstraint': '(max_skew=0, topology_key="",'
                                          ' when_unsatisfiable="")',
            'V1LimitRangeItem': '(type="")'
        }
        whole_kind = 'V1' + kind
        if whole_kind in must_param.keys():
            k8s_obj = eval('client.V1' + kind + must_param.get(whole_kind))
        else:
            k8s_obj = eval('client.V1' + kind + '()')
        self._init_k8s_obj(k8s_obj, file_content_dict, must_param)
        return k8s_obj

    def _get_k8s_obj_from_file_content_dict(self, file_content_dict,
                                            namespace=None):
        k8s_obj = {}
        kind = file_content_dict.get('kind', '')
        try:
            k8s_obj['object'] = self._create_k8s_object(
                kind, file_content_dict)
        except Exception as e:
            if isinstance(e, client.rest.ApiException):
                msg = '{kind} create failure. Reason={reason}'.format(
                    kind=file_content_dict.get('kind', ''), reason=e.body)
            else:
                msg = '{kind} create failure. Reason={reason}'.format(
                    kind=file_content_dict.get('kind', ''), reason=e)
            LOG.error(msg)
            raise exceptions.InitApiFalse(error=msg)

        k8s_obj['namespace'] = namespace
        if k8s_obj['object'].metadata:
            k8s_obj['object'].metadata.namespace = namespace

        return k8s_obj

    def get_k8s_objs_from_yaml(self, artifact_files, vnf_package_path,
                               namespace=None):
        k8s_objs = []
        for artifact_file in artifact_files:
            if ((urlparse(artifact_file).scheme == 'file') or
                    (bool(urlparse(artifact_file).scheme) and
                     bool(urlparse(artifact_file).netloc))):
                file_content = urllib2.urlopen(artifact_file).read()
            else:
                artifact_file_path = os.path.join(
                    vnf_package_path, artifact_file)
                with open(artifact_file_path, 'r') as f:
                    file_content = f.read()
            file_content_dicts = list(yaml.safe_load_all(file_content))
            for file_content_dict in file_content_dicts:
                k8s_obj = self._get_k8s_obj_from_file_content_dict(
                    file_content_dict, namespace)
                k8s_objs.append(k8s_obj)
        return k8s_objs

    def get_k8s_objs_from_manifest(self, mf_content, namespace):
        mkobj_kind_list = [
            "Pod",
            "Service",
            "PersistentVolumeClaim",
            "Namespace",
            "Node",
            "PersistentVolume",
            "DaemonSet",
            "Deployment",
            "ReplicaSet",
            "StatefulSet",
            "Job"
        ]
        k8s_objs = []
        mf_content_dicts = list(yaml.safe_load_all(mf_content))
        for mf_content_dict in mf_content_dicts:
            kind = mf_content_dict.get('kind', '')
            if kind in mkobj_kind_list:
                k8s_obj = self._get_k8s_obj_from_file_content_dict(
                    file_content_dict=mf_content_dict,
                    namespace=namespace)
                k8s_objs.append(k8s_obj)
        return k8s_objs

    def _select_k8s_client_and_api(
            self, kind, namespace, api_version, body):
        k8s_client_obj = self.k8s_client_dict[api_version]
        if 'namespaced' in self.method_value[kind]:
            response = getattr(k8s_client_obj, self.method_value.get(kind))(
                namespace=namespace, body=body
            )
        else:
            response = getattr(k8s_client_obj, self.method_value.get(kind))(
                body=body
            )
        return response

    def deploy(self, kubernetes_objects):
        """Deploy Kubernetes objects on Kubernetes VIM and return

        a list name of services
        """
        deployment_names = list()
        namespace = kubernetes_objects.get('namespace')
        k8s_objects = kubernetes_objects.get('objects')

        for k8s_object in k8s_objects:
            object_type = k8s_object.kind

            if object_type == 'ConfigMap':
                self.core_v1_api_client.create_namespaced_config_map(
                    namespace=namespace,
                    body=k8s_object)
                LOG.debug('Successfully created ConfigMap %s',
                          k8s_object.metadata.name)
            elif object_type == 'Deployment':
                self.app_v1_api_client.create_namespaced_deployment(
                    namespace=namespace,
                    body=k8s_object)
                LOG.debug('Successfully created Deployment %s',
                          k8s_object.metadata.name)
            elif object_type == 'HorizontalPodAutoscaler':
                self.scaling_api_client. \
                    create_namespaced_horizontal_pod_autoscaler(
                        namespace=namespace,
                        body=k8s_object)
                LOG.debug('Successfully created Horizontal Pod Autoscaler %s',
                          k8s_object.metadata.name)
            elif object_type == 'Service':
                self.core_v1_api_client.create_namespaced_service(
                    namespace=namespace,
                    body=k8s_object)
                LOG.debug('Successfully created Service %s',
                          k8s_object.metadata.name)
                deployment_names.append(namespace)
                deployment_names.append(k8s_object.metadata.name)

        # return a string that contains all deployment namespace and names
        # for tracking resources pattern:
        # namespace1,deployment1,namespace2,deployment2,namespace3,deployment3
        return ",".join(deployment_names)

    def deploy_k8s(self, kubernetes_objects):
        """Deploy kubernetes

        Deploy Kubernetes objects on Kubernetes VIM and
        return a list name of services
        """
        kubernetes_objects = self._sort_k8s_obj(kubernetes_objects)
        new_k8s_objs = list()
        for kubernetes_object in kubernetes_objects:
            namespace = kubernetes_object.get('namespace', '')
            kind = kubernetes_object.get('object', '').kind
            api_version = kubernetes_object.get('object', '').api_version
            body = kubernetes_object.get('object', '')
            if kubernetes_object.get('object', '').metadata:
                name = kubernetes_object.get('object', '').metadata.name
            else:
                name = ''
            try:
                LOG.debug("{kind} begin create.".format(kind=kind))
                self._select_k8s_client_and_api(
                    kind, namespace, api_version, body)
                kubernetes_object['status'] = 'Creating'
            except Exception as e:
                if isinstance(e, client.rest.ApiException):
                    kubernetes_object['status'] = 'creating_failed'
                    msg = '''The request to create a resource failed.
                    namespace: {namespace}, name: {name},kind: {kind},
                    Reason: {exception}'''.format(
                        namespace=namespace, name=name, kind=kind,
                        exception=e.body
                    )
                else:
                    kubernetes_object['status'] = 'creating_failed'
                    msg = '''The request to create a resource failed.
                    namespace: {namespace}, name: {name},kind: {kind},
                    Reason: {exception}'''.format(
                        namespace=namespace, name=name, kind=kind,
                        exception=e
                    )
                LOG.error(msg)
                raise exceptions.CreateApiFalse(error=msg)
            new_k8s_objs.append(kubernetes_object)
        return new_k8s_objs

    def _get_lower_case_name(self, name):
        name = name.strip()
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

    def _init_k8s_obj(self, obj, content, must_param):
        for key, value in content.items():
            param_value = self._get_lower_case_name(key)
            if hasattr(obj, param_value) and \
                    not isinstance(value, dict) and \
                    not isinstance(value, list):
                setattr(obj, param_value, value)
            elif isinstance(value, dict):
                obj_name = obj.openapi_types.get(param_value)
                if obj_name == 'dict(str, str)':
                    setattr(obj, param_value, value)
                else:
                    if obj_name in must_param.keys():
                        rely_obj = eval('client.' + obj_name +
                                       must_param.get(obj_name))
                    else:
                        rely_obj = eval('client.' + obj_name + '()')
                    self._init_k8s_obj(rely_obj, value, must_param)
                    setattr(obj, param_value, rely_obj)
            elif isinstance(value, list):
                obj_name = obj.openapi_types.get(param_value)
                if obj_name == 'list[str]':
                    setattr(obj, param_value, value)
                else:
                    rely_objs = []
                    rely_obj_name = \
                        re.findall(r".*\[([^\[\]]*)\].*", obj_name)[0]
                    for v in value:
                        if rely_obj_name in must_param.keys():
                            rely_obj = eval('client.' + rely_obj_name +
                                           must_param.get(rely_obj_name))
                        else:
                            rely_obj = \
                                eval('client.' + rely_obj_name + '()')
                        self._init_k8s_obj(rely_obj, v, must_param)
                        rely_objs.append(rely_obj)
                    setattr(obj, param_value, rely_objs)

    def _sort_k8s_obj(self, k8s_objs):
        pos = 0
        objs = k8s_objs
        sorted_k8s_objs = list()
        for sort_index, kind in enumerate(self.RESOURCE_CREATION_ORDER):
            for obj_index, obj in enumerate(objs):
                if obj["object"].kind == kind:
                    sorted_k8s_objs.append(objs.pop(obj_index))
            if sort_index == OTHER_RESOURCE_SORT_POSITION:
                pos = len(sorted_k8s_objs)
        for obj in objs:
            sorted_k8s_objs.insert(pos, obj)

        return sorted_k8s_objs

    def get_object_meta(self, content):
        must_param = {}
        v1_object_meta = client.V1ObjectMeta()
        self._init_k8s_obj(v1_object_meta, content, must_param)
        return v1_object_meta

    # config_labels configures label
    def config_labels(self, deployment_name=None, scaling_name=None):
        label = dict()
        if deployment_name:
            label.update({"selector": deployment_name})
        if scaling_name:
            label.update({"scaling_name": scaling_name})
        return label

    # Init resource requirement for container
    def init_resource_requirements(self, container):
        limits = dict()
        requests = dict()
        if container.num_cpus:
            limits.update({'cpu': container.num_cpus})
            requests.update({'cpu': container.num_cpus})
        if container.mem_size:
            limits.update({'memory': container.mem_size})
            requests.update({'memory': container.mem_size})
        return client.V1ResourceRequirements(limits=limits,
                                             requests=requests)

    def init_envs(self, container_props, name):
        config = container_props.config
        config_dict = self.pre_process_config(config)
        configmap_name = name

        list_envs = []
        for key in config_dict:
            config_map_ref = client.V1ConfigMapKeySelector(
                key=key,
                name=configmap_name)
            env_var = client.V1EnvVarSource(
                config_map_key_ref=config_map_ref)
            env_object = client.V1EnvVar(
                name=key,
                value_from=env_var)
            list_envs.append(env_object)
        return list_envs

    # Init container object
    def init_containers(self, container_props, limit_resource, name):
        list_env_var = self.init_envs(container_props, name)
        container_name = self.pre_process_name(container_props.name)
        list_container_port = list()
        if container_props.ports:
            for container_port in container_props.ports:
                port = int(container_port)
                cport = client.V1ContainerPort(container_port=port)
                list_container_port.append(cport)
        container = client.V1Container(
            name=container_name,
            image=container_props.image,
            ports=list_container_port,
            resources=limit_resource,
            command=container_props.command,
            args=container_props.args,
            env=list_env_var,
            image_pull_policy="IfNotPresent")
        return container

    # init_deployment initializes Kubernetes Pod object
    def init_deployment(self, tosca_kube_obj, kube_obj_name):
        """Instantiate the deployment object"""

        deployment_name = kube_obj_name
        # Create a list of container, which made a Pod
        containers = list()
        for container_prop in tosca_kube_obj.containers:
            limit_resource = self.init_resource_requirements(container_prop)
            container = self.init_containers(
                container_props=container_prop,
                limit_resource=limit_resource,
                name=deployment_name)
            containers.append(container)

        # Make a label with pattern {"selector": "deployment_name"}
        if tosca_kube_obj.scaling_object:
            scaling_name = tosca_kube_obj.scaling_object.scaling_name
            update_label = self.config_labels(deployment_name=deployment_name,
                                              scaling_name=scaling_name)
        else:
            update_label = self.config_labels(deployment_name=deployment_name)
        if tosca_kube_obj.labels:
            if 'selector' in update_label:
                del update_label['selector']
            update_label.update(tosca_kube_obj.labels)
        labels = update_label

        # Create and configure a spec section
        pod_template = client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(
                labels=labels, annotations=tosca_kube_obj.annotations),
            spec=client.V1PodSpec(containers=containers))
        # Create the specification of deployment
        label_selector = client.V1LabelSelector(match_labels=labels)
        deployment_spec = client.V1DeploymentSpec(
            template=pod_template, selector=label_selector)
        metadata = client.V1ObjectMeta(name=deployment_name, labels=labels)

        # Instantiate the deployment object
        deployment = client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=metadata,
            spec=deployment_spec)
        return deployment

    # init_hpa initializes Kubernetes Horizon Pod Auto-scaling object
    def init_hpa(self, tosca_kube_obj, kube_obj_name):
        scaling_props = tosca_kube_obj.scaling_object
        hpa = None
        if scaling_props:
            min_replicas = scaling_props.min_replicas
            max_replicas = scaling_props.max_replicas
            cpu_util = scaling_props.target_cpu_utilization_percentage
            deployment_name = kube_obj_name

            # Create target Deployment object
            target = client.V1CrossVersionObjectReference(
                api_version="apps/v1",
                kind="Deployment",
                name=deployment_name)
            # Create the specification of horizon pod auto-scaling
            hpa_spec = client.V1HorizontalPodAutoscalerSpec(
                min_replicas=min_replicas,
                max_replicas=max_replicas,
                target_cpu_utilization_percentage=cpu_util,
                scale_target_ref=target)
            metadata = client.V1ObjectMeta(name=deployment_name)
            # Create Horizon Pod Auto-Scaling
            hpa = client.V1HorizontalPodAutoscaler(
                api_version="autoscaling/v1",
                kind="HorizontalPodAutoscaler",
                spec=hpa_spec,
                metadata=metadata)
        return hpa

    # init_service initializes Kubernetes service object
    def init_service(self, tosca_kube_obj, kube_obj_name):
        list_service_port = list()
        service_label = tosca_kube_obj.labels
        for port in tosca_kube_obj.mapping_ports:
            if COLON_CHARACTER in port:
                ports = port.split(COLON_CHARACTER)
                published_port = int(ports[0])
                target_port = int(ports[1])
            else:
                target_port = published_port = int(port)
            service_port = client.V1ServicePort(
                name=str(published_port),
                port=published_port,
                target_port=target_port)
            list_service_port.append(service_port)

        deployment_name = kube_obj_name
        selector_by_name = self.config_labels(deployment_name)
        if tosca_kube_obj.labels:
            selectors = tosca_kube_obj.labels.copy()
        else:
            selectors = selector_by_name
        if tosca_kube_obj.mgmt_connection_point:
            service_label['management_connection'] = 'True'
        if tosca_kube_obj.network_name:
            service_label['network_name'] = tosca_kube_obj.network_name
        service_label['vdu_name'] = tosca_kube_obj.name

        metadata = client.V1ObjectMeta(name=deployment_name,
                                       labels=service_label)
        if tosca_kube_obj.service_type:
            service_type = tosca_kube_obj.service_type
        else:
            service_type = None
        service_spec = client.V1ServiceSpec(
            selector=selectors,
            ports=list_service_port,
            type=service_type)

        service = client.V1Service(
            api_version="v1",
            kind="Service",
            spec=service_spec,
            metadata=metadata)
        return service

    # init_config_map initializes Kubernetes ConfigMap object
    def init_configmap(self, container_props, kube_obj_name):
        config_map = None
        if container_props.config:
            configmap_name = kube_obj_name
            metadata = client.V1ObjectMeta(name=configmap_name)
            config_dict = self.pre_process_config(container_props.config)
            config_map = client.V1ConfigMap(
                api_version="v1",
                kind="ConfigMap",
                data=config_dict,
                metadata=metadata)
        return config_map

    def pre_process_name(self, name):
        # replace '_' by '-' to meet Kubernetes' requirement
        new_name = name.replace(DASH_CHARACTER, HYPHEN_CHARACTER).lower()
        return new_name

    def pre_process_config(self, config):
        # Split by separating lines
        config_dict = {}
        if config:
            configs = config.split(NEWLINE_CHARACTER)
            for config_item in configs:
                # Ignore if config_item is null
                if config_item:
                    # Strip all types of white-space characters
                    config_item = config_item.replace(
                        WHITE_SPACE_CHARACTER,
                        NON_WHITE_SPACE_CHARACTER)
                    config_prop = config_item.split(COLON_CHARACTER)
                    config_dict[config_prop[0]] = config_prop[1]
            # config_dict has the pattern such as
            # {'param1': 'key1', 'param0': 'key0'}
        return config_dict



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from tacker._i18n import _
from tacker.common.container import kubernetes_utils
from tacker.common import log
from tacker.common import utils
from tacker import context as t_context
from tacker.extensions import nfvo
from tacker.keymgr import API as KEYMGR_API
from tacker.nfvo.drivers.vim import abstract_vim_driver

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

KUBERNETES_OPTS = [
    cfg.BoolOpt('use_barbican', default=True,
                help=_('Use barbican to encrypt vim password if True'
                       ', save vim credentials in local file system'
                       ' if False'))
]
cfg.CONF.register_opts(KUBERNETES_OPTS, 'k8s_vim')


def config_opts():
    return [('k8s_vim', KUBERNETES_OPTS)]


class Kubernetes_Driver(abstract_vim_driver.VimAbstractDriver):
    """Driver for Kubernetes VIM

    """

    def __init__(self):
        self.kubernetes = kubernetes_utils.KubernetesHTTPAPI()

    def get_type(self):
        return 'kubernetes'

    def get_name(self):
        return 'Kubernetes VIM Driver'

    def get_description(self):
        return 'Kubernetes VIM Driver'

    def authenticate_vim(self, vim_obj):
        """Validate VIM auth attributes

        """
        auth_cred, file_descriptor = self._get_auth_creds(vim_obj)
        self._validate_vim(auth_cred, file_descriptor)
        self.clean_authenticate_vim(auth_cred, file_descriptor)

    def _get_auth_creds(self, vim_obj):
        auth_cred = vim_obj['auth_cred']
        file_descriptor = self._create_ssl_ca_file(auth_cred)
        auth_cred['auth_url'] = vim_obj['auth_url']
        if ('username' not in auth_cred) and ('password' not in auth_cred):
            auth_cred['username'] = 'None'
            auth_cred['password'] = None
        return auth_cred, file_descriptor

    def _create_ssl_ca_file(self, auth_cred):
        ca_cert = utils.none_from_string(auth_cred.get('ssl_ca_cert'))
        if ca_cert:
            file_descriptor, file_path = \
                self.kubernetes.create_ca_cert_tmp_file(ca_cert)
            auth_cred['ca_cert_file'] = file_path
            return file_descriptor
        else:
            return None

    def _validate_vim(self, auth, file_descriptor):
        # If Tacker can get k8s_info, Kubernetes authentication is valid
        # if not, it is invalid
        auth_dict = dict(auth)
        try:
            core_api_client = \
                self.kubernetes.get_core_api_client(auth_dict)
            k8s_info = core_api_client.get_api_versions()
            LOG.info(k8s_info)
        except Exception as e:
            LOG.info('VIM Kubernetes authentication is wrong.')
            # delete temp file
            self.clean_authenticate_vim(auth_dict, file_descriptor)
            raise nfvo.VimUnauthorizedException(message=str(e))

    def _find_regions(self, core_v1_api_client):
        list_namespaces = core_v1_api_client.list_namespace()
        namespaces = [namespace.metadata.name
                      for namespace in list_namespaces.items]
        return namespaces

    def discover_placement_attr(self, vim_obj):
        """Fetch VIM placement information

        Attributes can include regions, AZ, namespaces.
        """
        # in Kubernetes environment, user can deploy resource
        # on specific namespace
        auth_cred, file_descriptor = self._get_auth_creds(vim_obj)
        core_v1_api_client = \
            self.kubernetes.get_core_v1_api_client(auth_cred)
        namespace_list = self._find_regions(core_v1_api_client)
        self.clean_authenticate_vim(auth_cred, file_descriptor)
        vim_obj['placement_attr'] = {'regions': namespace_list}
        return vim_obj

    def clean_authenticate_vim(self, vim_auth, file_descriptor):
        # remove ca_cert_file from vim_obj if it exists
        # close and delete temp ca_cert_file
        if file_descriptor is not None:
            file_path = vim_auth.pop('ca_cert_file')
            self.kubernetes.close_tmp_file(file_descriptor, file_path)

    @log.log
    def register_vim(self, vim_obj):
        """Validate Kubernetes VIM."""
        if 'key_type' in vim_obj['auth_cred']:
            vim_obj['auth_cred'].pop('key_type')
        if 'secret_uuid' in vim_obj['auth_cred']:
            vim_obj['auth_cred'].pop('secret_uuid')
        self.authenticate_vim(vim_obj)
        self.discover_placement_attr(vim_obj)
        self.encode_vim_auth(vim_obj['id'],
                             vim_obj['auth_cred'])
        LOG.debug('VIM registration completed for %s', vim_obj)

    @log.log
    def deregister_vim(self, vim_obj):
        """Deregister Kubernetes VIM from NFVO

        Delete VIM keys from file system
        """
        self.delete_vim_auth(vim_obj['id'],
                             vim_obj['auth_cred'])

    @log.log
    def delete_vim_auth(self, vim_id, auth):
        """Delete kubernetes vim information

        Delete vim key stored in file system
        """
        if 'secret_uuid' in auth:
            # Delete secret id of barbican
            LOG.debug('Attempting to delete key for vim id %s',
                      vim_id)
            if auth.get('key_type') == 'barbican_key':
                try:
                    k_context = \
                        t_context.generate_tacker_service_context()
                    keystone_conf = CONF.keystone_authtoken
                    secret_uuid = auth['secret_uuid']
                    keymgr_api = KEYMGR_API(keystone_conf.auth_url)
                    keymgr_api.delete(k_context, secret_uuid)
                    LOG.debug('VIM key deleted successfully for vim %s',
                              vim_id)
                except Exception as exception:
                    LOG.warning('VIM key deletion failed for vim %s due to %s',
                                vim_id,
                                exception)
                    raise
            else:
                raise nfvo.VimEncryptKeyError(vim_id=vim_id)

    @log.log
    def encode_vim_auth(self, vim_id, auth):
        """Encode VIM credentials

        Store VIM auth using fernet key encryption
        """
        fernet_key, fernet_obj = self.kubernetes.create_fernet_key()
        if ('password' in auth) and (auth['password'] is not None):
            encoded_auth = fernet_obj.encrypt(
                auth['password'].encode('utf-8'))
            auth['password'] = encoded_auth
        if 'bearer_token' in auth:
            encoded_auth = fernet_obj.encrypt(
                auth['bearer_token'].encode('utf-8'))
            auth['bearer_token'] = encoded_auth
        if utils.none_from_string(auth.get('ssl_ca_cert')):
            encoded_auth = fernet_obj.encrypt(
                auth['ssl_ca_cert'].encode('utf-8'))
            auth['ssl_ca_cert'] = encoded_auth

        if CONF.k8s_vim.use_barbican:
            try:
                k_context = t_context.generate_tacker_service_context()
                keystone_conf = CONF.keystone_authtoken
                keymgr_api = KEYMGR_API(keystone_conf.auth_url)
                secret_uuid = keymgr_api.store(k_context, fernet_key)

                auth['key_type'] = 'barbican_key'
                auth['secret_uuid'] = secret_uuid
                LOG.debug('VIM auth successfully stored for vim %s',
                          vim_id)
            except Exception as exception:
                LOG.warning('VIM key creation failed for vim %s due to %s',
                            vim_id,
                            exception)
                raise
        else:
            raise nfvo.VimEncryptKeyError(vim_id=vim_id)

    def get_vim_resource_id(self):
        # TODO(phuoc): will update which vim resource need to get
        pass
